import logging
from collections import Counter
from dataclasses import replace
from typing import Any, List, Optional

from deprecated import deprecated

import joblib
import onnxruntime
from datasets import Dataset, concatenate_datasets
from sentence_transformers import SentenceTransformer
import numpy as np
from setfit import SetFitModel, Trainer, TrainingArguments
from sklearn.metrics import classification_report, confusion_matrix
from transformers import AutoTokenizer, set_seed
from transformers.integrations.integration_utils import ClearMLCallback
from accelerate.utils import find_executable_batch_size
from open_autonlu.progress_reporter import (
    ProgressReporter,
    TrainerProgressReporterCallback,
)

from ..constants import (
    TRAINING_STAGE,
)
from .configs import SetFitMethodConfig
from .data_types import (
    ClassificationEvaluationResult,
    SaveFormat,
    TextClassifierOutput,
    TextLabel,
    TrainingArtifactInfo,
)
from .method import SequenceClassificationMethod
from ..methods.constants import OOS_LABEL
from ..methods.data_types import OodMethod
from .models.onnx_wrappers import SetFitWrapper
from .onnx_inference_manager import (
    AbstractONNXInferenceManager,
    GeneralSequenceClassifierInferenceManager,
)
from .utils import GenericSequenceClassifierONNXExportMixin
from .constants import BATCH_SIZE
from .utils import batch as batcher, GibberishDatasetGenerator

log = logging.getLogger(__name__)


class SetFitMethodBase(
    SequenceClassificationMethod, GenericSequenceClassifierONNXExportMixin
):
    """Base class for SetFit few-shot text classification.

    SetFit is a few-shot learning method that finetunes a sentence transformer and trains a classification head
    on the resulting embeddings.

    Attributes:
        config_cls: Configuration class for this method.
        stages: List of training stages for progress reporting.
    """

    config_cls = SetFitMethodConfig
    stages = [TRAINING_STAGE]

    def __init__(
        self,
        config: Optional[SetFitMethodConfig] = None,
        progress_reporter: Optional[ProgressReporter] = None,
    ) -> None:
        """Initialize the SetFit method.

        Args:
            config: Configuration object with model and training parameters.
                If None, uses default SetFitMethodConfig.
            progress_reporter: Reporter for tracking training progress.
                If None, creates a default reporter.
        """
        super().__init__(config)
        if config is None:
            config = SetFitMethodConfig()
        if progress_reporter is None:
            progress_reporter = ProgressReporter()
        self.progress_reporter = progress_reporter

        if config.model_name_or_path is not None:
            self.model = SetFitModel.from_pretrained(config.model_name_or_path)
        else:
            self.model = None
        self.ood_method = config.ood_method
        self.body_lr = config.body_lr
        self.num_iterations = config.num_iterations
        self.model.model_body.tokenizer.model_max_length = config.max_seq_length  # hardcoded for russian encoders, that lack their own model_max_length parameter
        self.seed = config.seed
        self.batch_size = config.batch_size
        set_seed(config.seed)

    @property
    def id2label(self):
        if getattr(self, "model"):
            return self.model.config.id2label
        else:
            return self._id2label

    @id2label.setter
    def id2label(self, value):
        self._id2label = value

    def _generate_synthetic_oos(self, num_train_rows: int) -> Dataset:
        generator = GibberishDatasetGenerator(self.seed)
        synthetic_texts = generator(num_rows=num_train_rows)
        synthetic_dataset = Dataset.from_dict(
            {"text": synthetic_texts, "label": [OOS_LABEL] * num_train_rows}
        )
        return synthetic_dataset

    def train(self, dataset: Dataset) -> TrainingArtifactInfo:
        """Train the SetFit model on the provided dataset.

        Args:
            dataset: Training dataset with 'text' and 'label' columns.

        Returns:
            TrainingArtifactInfo with training arguments and label list.
        """

        self.progress_reporter.stages = self.stages
        self.progress_reporter(TRAINING_STAGE, 0.0)
        fit_with_auto_batch_size = find_executable_batch_size(
            self._fit_with_auto_batch_size, starting_batch_size=self.batch_size
        )
        label_counts = Counter(dataset["label"])
        ood_count = label_counts.get(OOS_LABEL, 0)
        p95 = int(np.percentile(list(label_counts.values()), 95))
        if self.ood_method == OodMethod.NONE:
            if ood_count > 0:
                synthetic_texts_ds = self._generate_synthetic_oos(p95 - ood_count)
                dataset = concatenate_datasets([dataset, synthetic_texts_ds])
        elif self.ood_method == OodMethod.LOGIT:
            if ood_count == 0:
                synthetic_texts_ds = self._generate_synthetic_oos(p95)
                dataset = concatenate_datasets([dataset, synthetic_texts_ds])
            elif ood_count < p95:
                synthetic_texts_ds = self._generate_synthetic_oos(p95 - ood_count)
                dataset = concatenate_datasets([dataset, synthetic_texts_ds])
        trainer = fit_with_auto_batch_size(dataset=dataset)
        # Model swapping
        self.base_setfit_instance = self.model
        model_pooler = self.model.model_body._modules["1"]

        self.model = SetFitWrapper(
            self.model.model_body._modules["0"].auto_model,
            model_pooler,
            "2" in self.model.model_body._modules,
            self.model.model_head,
        ).eval()
        return TrainingArtifactInfo(trainer.args.to_dict(), self.label_list)

    def _fit_with_auto_batch_size(self, batch_size: int, dataset: Dataset) -> Trainer:
        trainer = Trainer(
            model=self.model,
            train_dataset=dataset,
            args=TrainingArguments(
                batch_size=batch_size,
                body_learning_rate=self.body_lr,
                report_to="none",
                seed=self.seed,
            ),
        )
        # argument `report_to="none"` is ignored during instantiation of BCSentenceTransformersTrainer
        # inside of Trainer, so we have to remove all integration callbacks manually
        if hasattr(trainer, "st_trainer"):
            if "clearml" not in trainer.args.report_to:
                trainer.st_trainer.remove_callback(ClearMLCallback)
            trainer.st_trainer.add_callback(
                TrainerProgressReporterCallback(self.progress_reporter, TRAINING_STAGE)
            )
        if self.num_iterations is not None:
            trainer.args = replace(trainer.args, num_iterations=self.num_iterations)
        trainer.train()
        return trainer

    def test(
        self, dataset: Dataset, inscope_only: bool = False
    ) -> ClassificationEvaluationResult:
        """Evaluate the model on a test dataset.

        Args:
            dataset: Test dataset with 'text' and 'label' columns.
            inscope_only: If True, exclude OOS predictions and labels from
                evaluation, replacing OOS predictions with the most probable
                in-scope label.

        Returns:
            ClassificationEvaluationResult with confusion matrix,
            classification report, and macro F1 score.
        """
        predictions = self.predict(dataset["text"], batch_size=BATCH_SIZE)
        labels_hat = [pred.most_probable.label for pred in predictions]

        if inscope_only:
            for i, pred in enumerate(predictions):
                if pred.most_probable.label == OOS_LABEL:
                    inscope_hypotheses = [
                        h for h in pred.hypotheses if h.label != OOS_LABEL
                    ]
                    if inscope_hypotheses:
                        inscope_hypotheses.sort(key=lambda x: x.score, reverse=True)
                        labels_hat[i] = inscope_hypotheses[0].label

            original_labels = dataset["label"]
            mask = [label != OOS_LABEL for label in original_labels]
            dataset = dataset.filter(lambda example, idx: mask[idx], with_indices=True)
            labels_hat = [label for label, keep in zip(labels_hat, mask) if keep]
            inscope_labels = [label for label in self.label_list if label != OOS_LABEL]
        else:
            inscope_labels = None

        labels_for_conf_matrix = inscope_labels if inscope_only else self.label_list
        ds_label_set = set(dataset.unique("label"))
        if inscope_only and OOS_LABEL in ds_label_set:
            ds_label_set.remove(OOS_LABEL)
        model_label_set = set(labels_for_conf_matrix)
        extra_in_test = ds_label_set - model_label_set
        if extra_in_test:
            log.warning(
                "Test dataset contains labels not seen during training: %s",
                sorted(extra_in_test),
            )

        clf_report = classification_report(
            dataset["label"],
            labels_hat,
            labels=sorted(ds_label_set),
            output_dict=True,
            zero_division=0,
        )
        # Mitigating the issue with missing labels in model label set
        # while retaining the order of labels in the confusion matrix
        if ds_label_set != model_label_set:
            labels_for_conf_matrix = labels_for_conf_matrix + list(
                ds_label_set - model_label_set
            )

        conf_matrix = confusion_matrix(
            dataset["label"], labels_hat, labels=labels_for_conf_matrix
        )
        return ClassificationEvaluationResult(
            conf_matrix, clf_report, clf_report["macro avg"]["f1-score"]
        )

    def _predict(
        self,
        batch_size: int,
        texts: List[str],
    ) -> list[TextClassifierOutput]:
        if isinstance(texts, str):
            texts = [texts]

        logits = self.base_setfit_instance.predict_proba(
            texts, batch_size=batch_size, as_numpy=True
        )
        labels = self._extract_labels(logits, self.id2label)
        scores = self._extract_scores(logits)
        output = [
            TextClassifierOutput(text=text, hypotheses=[TextLabel(label, score)])
            for label, score, text in zip(labels, scores, texts)
        ]
        return output

    def save(
        self, path_to_package: str, checkpoint_format: SaveFormat = SaveFormat.TORCH
    ) -> None:
        """Save the trained model to disk.

        Args:
            path_to_package: Directory path for saving the model.
            checkpoint_format: Format to save in. TORCH saves the full
                SetFit model; ONNX exports an optimized inference model.
        """
        if checkpoint_format == SaveFormat.TORCH:
            self.base_setfit_instance.save_pretrained(path_to_package)
        else:
            self.tokenizer = self.base_setfit_instance.model_body.tokenizer
            self.export_onnx(path_to_package)
            self.tokenizer.save_pretrained(path_to_package)
        log.info("Model is saved at " + path_to_package)

    @classmethod
    def load(cls, path_to_package: str):
        """Load a trained SetFit model from disk.

        Args:
            path_to_package: Directory path containing the saved model.

        Returns:
            Initialized SetFitMethodBase instance with loaded model.
        """
        self = cls()  # TODO: add config loading
        model_body = SentenceTransformer(path_to_package)
        model_head = joblib.load(path_to_package + "/model_head.pkl")
        self.model = SetFitModel(model_body=model_body, model_head=model_head)
        self.base_setfit_instance = self.model
        self.model = SetFitWrapper(
            model_body._modules["0"].auto_model,
            model_body._modules["1"],
            "2" in model_body._modules,
            model_head,
        ).eval()
        return self


@deprecated(
    version="0.3.5",
    reason="This class is outdated, use GeneralSequenceClassifierInferenceManager instead.",
)
class SetFitInferenceManager(AbstractONNXInferenceManager):
    def __init__(self, path_to_package: str) -> None:
        super().__init__(path_to_package)
        self.tokenizer = AutoTokenizer.from_pretrained(path_to_package)
        self.session = onnxruntime.InferenceSession(
            path_to_package + "/model.onnx",
            providers=[self.ort_provider, "CPUExecutionProvider"],
        )

    def predict(
        self,
        texts: List[str],
        batch_size: int = BATCH_SIZE,
        auto_find_batch_size: bool = False,
        return_all_hypotheses: bool = False,
    ):
        if return_all_hypotheses:
            log.debug(
                "return_all_hypotheses is not supported for torch-based inference for now and will be ignored."
            )
        if auto_find_batch_size:
            return find_executable_batch_size(
                self._predict, starting_batch_size=batch_size
            )(texts)  # type: ignore
        return self._predict(batch_size, texts)

    def _predict(self, batch_size: int, texts: List[str]) -> List[Any]:
        all_labels, all_scores = [], []
        for text_batch in batcher(texts, batch_size):
            inputs = self.tokenizer(
                text_batch,
                padding=True,
                truncation=True,
                return_attention_mask=True,
                return_token_type_ids=True,
                return_tensors="np",
            )
            labels, logits = self.session.run(None, dict(inputs))
            scores = SetFitMethodBase._extract_scores(logits)
            all_labels.extend(labels)
            all_scores.extend(scores)
        return [
            {"label": label, "score": score}
            for label, score in zip(all_labels, all_scores)
        ]


class SetFitMethod(SetFitMethodBase):
    """SetFit method for few-shot text classification.

    Concrete implementation of SetFitMethodBase that uses the standard
    GeneralSequenceClassifierInferenceManager for ONNX inference.

    This is the recommended SetFit variant for production use.
    """

    onnx_inference_manager = GeneralSequenceClassifierInferenceManager



from typing import Callable, Optional

from transformers.trainer_callback import (
    TrainerCallback,
    TrainerControl,
    TrainerState,
    TrainingArguments,
)

from .data_types import ProgressCallbackPayload


class ProgressReporter:
    """
    A class for reporting progress during training.

    Attributes:
        progress_callback (Callable[[ProgressCallbackPayload], None]): A callback function to report progress.
        _stages (Optional[list]): A list of stages in the training process.
        state (ProgressCallbackPayload): The current state of the training progress.

    Methods:
        __call__(stage_name: str, stage_progress: float): Updates and reports the progress based on the current stage and its progress.
    """

    def __init__(
        self,
        progress_callback: Optional[Callable[[ProgressCallbackPayload], None]] = None,
    ) -> None:
        self._stages: Optional[list] = None
        self.is_passthrough = False
        if progress_callback is None:
            self.is_passthrough = True
        self.progress_callback = progress_callback
        self.state = ProgressCallbackPayload("", 0)

    @property
    def stages(self):
        return self._stages

    @stages.setter
    def stages(self, value):
        """
        Setter method for the stages attribute.
        Updates the list of stages and calculates the percentage of progress per stage.

        Args:
            value (list): A list of stage names in the training process.
        """
        self._stages = value
        self._percent_per_stage = 1 / len(value)

    def __call__(self, stage_name: str, stage_progress: float):
        """
        Updates and reports the progress based on the current stage and its progress.

        Requirements:
            - The stages attribute must be set before calling this method.
            - stage_name (str): The name of the current stage.
            - stage_progress (float): The progress within the current stage, ranging from 0 to 1.
        """
        if self.is_passthrough:
            return
        assert self._stages is not None, "Stage name list is not set"
        self.state.stage_name = stage_name
        self.state.percent = (
            self._percent_per_stage * self._stages.index(stage_name)
        ) + self._percent_per_stage * stage_progress
        self.progress_callback(self.state)


class TrainerProgressReporterCallback(TrainerCallback):
    """
    A custom trainer callback for reporting progress during training.

    This class extends the TrainerCallback and provides a way to report progress at different stages of the training process.
    It can be used in conjunction with ProgressReporter to update the progress and trigger callbacks.

    Attributes:
        progress_reporter (ProgressReporter): The instance of ProgressReporter responsible for updating and reporting progress.
        stage (str): The current stage name being processed.

    Methods:
        on_train_begin(args: TrainingArguments, state: TrainerState, control: TrainerControl): Called at the beginning of training. Updates the progress reporter with initial stage and progress.
        on_train_end(args: TrainingArguments, state: TrainerState, control: TrainerControl): Called when training is complete. Updates the progress reporter with final stage and progress.
        on_step_end(args: TrainingArguments, state: TrainerState, control: TrainerControl): Called at the end of each step. Updates the progress reporter with current stage and progress.
    """

    def __init__(self, progress_reporter: ProgressReporter, stage: str) -> None:
        """
        Initializes the TrainerProgressReporterCallback.

        Args:
            progress_reporter (ProgressReporter): The instance of ProgressReporter responsible for updating and reporting progress.
            stage (str): The current stage name being processed.
        """
        self.progress_reporter = progress_reporter
        self.stage = stage

    def on_train_begin(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Called at the beginning of training.

        Updates the progress reporter with initial stage and progress.
        """
        self.progress_reporter(self.stage, 0.0)

    def on_train_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Called when training is complete.

        Updates the progress reporter with final stage and progress.
        """
        self.progress_reporter(self.stage, 1.0)

    def on_step_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Called at the end of each step.

        Updates the progress reporter with current stage and progress.
        """
        self.progress_reporter(self.stage, state.global_step / state.max_steps)



from typing import Any, Dict, List, Tuple
from ...nervaluate import Evaluator
import numpy as np


def evaluate_entity_level(
    true_labels: List[List[str]], pred_labels: List[List[str]], entity_names: List[str]
) -> Tuple[Dict[str, Any], np.ndarray]:
    evaluator = Evaluator(true_labels, pred_labels, tags=entity_names, loader="list")
    overall, entity_type, _, _, confusion_matrix = evaluator.evaluate()
    result = {}
    for name in entity_names:
        metrics = entity_type[name]["strict"]
        result[name] = {
            "precision": metrics["precision"],
            "recall": metrics["recall"],
            "f1-score": metrics["f1"],
            "support": metrics["possible"],
        }
    overal_metrics = overall["strict"]
    result["macro avg"] = {
        "precision": overal_metrics["precision"],
        "recall": overal_metrics["recall"],
        "f1-score": overal_metrics["f1"],
        "support": overal_metrics["possible"],
    }
    return result, confusion_matrix



from dataclasses import dataclass
from typing import Optional
import os
from ..prompts import DEFAULT_SYSTEM_PROMPT


@dataclass
class LlmClientConfig:
    """Configuration for LLM API client.

    Used for data augmentation and synthetic test generation pipelines.
    Defaults are read from environment variables if available.

    Attributes:
        api_key: API key for authentication. Reads from MODEL_API_KEY env var.
        base_url: Base URL of the LLM API endpoint. Reads from BASE_URL env var.
        model_id: Model identifier to use. Reads from MODEL_ID env var.
        default_system_prompt: System prompt for the LLM.
        default_temperature: Sampling temperature (0.0-1.0).
        default_presence_penalty: Presence penalty for generation.
        default_frequency_penalty: Frequency penalty for generation.
        num_concurrent_requests: Maximum concurrent API requests.
    """

    api_key: str = os.environ.get("MODEL_API_KEY", "")
    base_url: str = os.environ.get("BASE_URL", "")
    model_id: str = os.environ.get("MODEL_ID", "")
    default_system_prompt: Optional[str] = DEFAULT_SYSTEM_PROMPT
    default_temperature: Optional[float] = 0.5
    default_presence_penalty: Optional[float] = 0.0
    default_frequency_penalty: Optional[float] = 0.1
    num_concurrent_requests: Optional[int] = 1



from abc import ABC, abstractmethod

import pandas as pd
from datasets import Dataset, DatasetDict
from typing import Optional

from ..constants import TRAIN_KEY, DEV_KEY, TEST_KEY
import logging

log = logging.getLogger(__name__)


class AbstractDataProvider(ABC):
    """Abstract base class for data providers.

    Data providers handle loading and preprocessing of training and
    evaluation datasets from various file formats.
    """

    def __init__(
        self,
        train_path: str,
        dev_path: Optional[str] = None,
        test_path: Optional[str] = None,
        language: str = "en",
    ) -> None:
        """Initialize the data provider.

        Args:
            train_path: Path to training data file.
            dev_path: Optional path to development/validation data.
            test_path: Optional path to test data.
            language: "ru" or "en" for tokenization.
        """
        self.train_path = train_path
        self.dev_path = dev_path
        self.test_path = test_path
        self.language = language

    def load(self) -> DatasetDict:
        """Load and preprocess data from configured paths.

        Returns:
            DatasetDict containing train, dev, and/or test splits.
        """
        data = self._load_data()
        data = self._postprocess(data)

        return data

    @abstractmethod
    def _load_data(self) -> DatasetDict:
        raise NotImplementedError()

    def _postprocess(self, data: DatasetDict) -> DatasetDict:
        return data


class SimpleDataProvider(AbstractDataProvider):
    """Data provider for CSV files with text and label columns.

    Loads CSV files containing 'text' and 'label' columns.
    Optionally supports 'anc_label' column for anchor-based methods.
    """

    def __init__(
        self,
        train_path: str,
        num_examples: Optional[int] = None,
        dev_path: Optional[str] = None,
        test_path: Optional[str] = None,
        language: str = "en",
    ) -> None:
        """Initialize the CSV data provider.

        Args:
            train_path: Path to training CSV file.
            num_examples: Optional limit on number of training examples.
            dev_path: Optional path to development CSV.
            test_path: Optional path to test CSV.
            language: "ru" or "en".
        """
        super().__init__(train_path, dev_path, test_path, language=language)
        self.num_examples = num_examples

    def _load_data(self) -> dict[str, pd.DataFrame]:
        splits = {}
        for split_name in [TRAIN_KEY, DEV_KEY, TEST_KEY]:
            if path := getattr(self, f"{split_name}_path"):
                df = pd.read_csv(path)
                selector = ["text", "label"]
                if split_name == TRAIN_KEY:
                    if "anc_label" in df.keys():
                        selector = ["text", "label", "anc_label"]
                    if self.num_examples is not None:
                        splits[split_name] = df[selector][: self.num_examples]
                        continue
                splits[split_name] = df[selector]
        return splits

    def _postprocess(self, data: dict[str, pd.DataFrame]) -> DatasetDict:
        """
        We remove all examples with OOS_LABEL from the train split, as they are not used for training.
        We remove all examples, that have their columns 'text' or 'label' empty and warn user about removed
        examples.
        """

        for split in data.keys():
            df = data[split]
            indices = df[
                df.label.isna() | df.text.isna() | (df.text == "") | (df.label == "")
            ].index
            if len(indices):
                data[split] = df.drop(index=indices)
                log.warning(
                    f"Some rows contain empty values in {split} split. Rows with indices {indices.to_list()} were removed to avoid any runtime errors."
                )
        return DatasetDict(
            {
                split: Dataset.from_pandas(df, preserve_index=False)
                for split, df in data.items()
            }
        )



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Union
import logging

from datasets import Dataset, DatasetDict
from iterstrat.ml_stratifiers import MultilabelStratifiedShuffleSplit
from sklearn.model_selection import train_test_split
import numpy as np
from accelerate.utils import find_executable_batch_size
from open_autonlu.progress_reporter import ProgressReporter
from open_autonlu.methods.constants import BATCH_SIZE


from .data_types import (
    ClassificationEvaluationResult,
    NERFormat,
    RetrievalEvaluationResult,
    TokenClassifierOutput,
    TextClassifierOutput,
    TrainingArtifactInfo,
    SaveFormat,
)
from .configs.base_config import BaseMethodConfig

log = logging.getLogger(__name__)


class Method(ABC):
    @abstractmethod
    def __init__(
        self,
        config: Optional[BaseMethodConfig] = None,
        progress_reporter: Optional[ProgressReporter] = None,
    ) -> None: ...

    @abstractmethod
    def train(self, dataset: Dataset) -> TrainingArtifactInfo:
        """
        Takes in a dataset, that contains train split
        with texts and labels.
        """
        ...

    @abstractmethod
    def test(
        self, dataset: Dataset
    ) -> Union[ClassificationEvaluationResult, RetrievalEvaluationResult]:
        """
        Takes in a dataset, that contains test split
        with texts and labels. Runs inference on texts and counts needed metrics.
        Returns a sklearn classification_report
        """
        ...

    @abstractmethod
    def save(
        self, path_to_package: str, checkpoint_format: SaveFormat = SaveFormat.TORCH
    ) -> None:
        """
        Serializes all the objects and files to a designated path for future recreation.
        """
        ...

    @classmethod
    @abstractmethod
    def load(cls, path_to_package: str):
        """
        Using files from path_to_package reconstructs its own exemplar
        """
        ...

    @staticmethod
    def _get_label_mappings(labels: List[str]):
        """
        Gets the label mappings from a list of labels
        """
        label2id = {label: l_id for l_id, label in enumerate(labels)}
        id2label = {l_id: label for l_id, label in enumerate(labels)}
        return label2id, id2label


class SequenceClassificationMethod(Method):
    @property
    def label_list(self) -> List[str]:
        return [
            value
            for _, value in sorted(self.id2label.items(), key=lambda item: item[0])
        ]

    @staticmethod
    def _extract_scores(logits) -> List[float]:
        return logits.max(-1).tolist()

    @staticmethod
    def _extract_labels(logits, id2label) -> List[str]:
        human_readable_labels = []
        for label_id in logits.argmax(-1):
            human_readable_labels.append(id2label[label_id])
        return human_readable_labels

    @abstractmethod
    def test(
        self, dataset: Dataset, inscope_only: bool = False
    ) -> ClassificationEvaluationResult:
        """
        Takes in a dataset, that contains test split
        with texts and labels. Runs inference on texts and counts needed metrics.
        Returns a sklearn classification_report.

        Args:
            dataset: Dataset with test data
            inscope_only: If True, exclude outOfScope labels from inference predictions,
                         replacing them with the most probable in-scope label
        """
        ...

    @abstractmethod
    def _predict(
        self,
        batch_size: int,
        texts: List[str],
    ) -> List[TextClassifierOutput]:
        """
        Takes in a collection of texts and runs inference on it. Returns a collection of string labels.
        """
        ...

    def predict(
        self,
        texts: List[str],
        batch_size: int = BATCH_SIZE,
        auto_find_batch_size: bool = False,
        return_all_hypotheses: bool = False,
    ) -> List[TextClassifierOutput]:
        """
        Takes in a collection of texts and runs inference on it. Returns a collection of string labels.
        """
        if return_all_hypotheses:
            log.debug(
                "return_all_hypotheses is not supported for torch-based inference for now and will be ignored."
            )
        if auto_find_batch_size:
            return find_executable_batch_size(
                self._predict, starting_batch_size=batch_size
            )(texts=texts)

        return self._predict(batch_size, texts)


class TokenClassificationMethod(Method):
    """
    This class of methods consumes (through .train)
    and yields (through .predict) the following data structures accordingly:
    .train(Dataset([{"text": "the text to analyze", "labels": ["B-entity", "I-entity", "O", "O"]}]))
    .predict("the text to analyze" or ["the text to analyze", ...]) -> List[TokenClassifierOutput]
    """

    @abstractmethod
    def predict(
        self,
        texts: Union[List[str], str],
        batch_size: int,
        output_format: NERFormat = NERFormat.OFFSETS,
    ) -> List[TokenClassifierOutput]:
        """
        Takes in a collection of texts and runs inference on it. Returns a collection of objects of type TokenClassifierOutput.
        """
        ...

    def train_test_split(
        self,
        dataset: Dataset,
        entity_names: List[str],
        test_size: float = 0.1,
        seed: int = 42,
        bio_tags_key: str = "labels",
    ) -> DatasetDict:
        indices = np.arange(dataset.num_rows, dtype=int)
        entity2id = {name: idx for idx, name in enumerate(entity_names)}
        bag_of_labels = self._get_bag_of_labels(dataset[bio_tags_key], entity2id)
        if bag_of_labels.shape[1] == 1:
            train_indices, test_indices = train_test_split(
                indices, test_size=test_size, random_state=seed, stratify=bag_of_labels
            )
        else:
            splitter = MultilabelStratifiedShuffleSplit(
                n_splits=1, test_size=test_size, random_state=seed
            )
            train_indices, test_indices = next(splitter.split(indices, bag_of_labels))
        return DatasetDict(
            {
                "train": dataset.select(train_indices),
                "test": dataset.select(test_indices),
            }
        )

    def _get_bag_of_labels(
        self, labels: List[List[str]], label2id: Dict[str, int]
    ) -> np.array:
        outputs = []
        for bio_tags in labels:
            label_names = list({tag[2:] for tag in bio_tags if tag != "O"})
            label_ids = [label2id[label] for label in label_names]
            bag_of_labels = np.zeros(len(label2id), dtype=int)
            bag_of_labels[label_ids] = 1
            outputs.append(bag_of_labels)
        return np.vstack(outputs)



"""
Module for data quality management
"""

__version__ = "0.5"

from .dynamic_finetuner import SeqClassDynamicTuner, TokenClassDynamicTuner
from .evaluator import DatasetEvaluator
from .sample_eval import Cartography, LabelAggregation, Retag, Uncertainty, VInfo

__all__ = [
    "DatasetEvaluator",
    "Cartography",
    "LabelAggregation",
    "Retag",
    "Uncertainty",
    "VInfo",
    "SeqClassDynamicTuner",
    "TokenClassDynamicTuner",
]



from dataclasses import dataclass
from .base_config import BaseMethodWithOODConfig
from .anc_setfit_config import AncSetFitConfig
from ..data_types import OodMethod


@dataclass
class AncSetFitOODConfig(AncSetFitConfig, BaseMethodWithOODConfig):
    """Configuration for Anchor-based SetFit with Out-of-Distribution detection.

    Extends AncSetFitConfig with OOD detection capabilities using Maximum Softmax
    Probability (MSP) by default. Combines anchor-based contrastive learning with
    the ability to detect inputs that don't belong to any known class.

    Warning:
        OOD detection does not perform well in few-shot settings with this method.
        Based on benchmarks, the OOD detection metrics are significantly degraded
        when training data is limited. For few-shot scenarios (under 5 samples per class),
        OOD detection is not turned off by default.

    Attributes:
        threshold_factor: Multiplier for OOD detection threshold. Higher values
            make detection more conservative (fewer samples marked as OOD).
        ood_method: OOD detection method to use. Defaults to MSP (Maximum Softmax
            Probability).
    """

    threshold_factor: float = 1.0
    ood_method: OodMethod = OodMethod.MSP_OOD



"""
Example: text classification pipeline.
"""

import torch

from open_autonlu.auto_classes import (
    TextClassificationTrainingPipeline,
)

TRAIN_PATH = "examples/test_data/noise_n_shot_data/snips_200_300_0.01_42_train.csv"
TEST_PATH = "examples/test_data/noise_n_shot_data/snips_200_300_0.01_42_test.csv"


def main():
    if not torch.cuda.is_available() and not torch.backends.mps.is_available():
        print("Warning: no GPU; training may be slow.")

    pipeline = TextClassificationTrainingPipeline(
        TRAIN_PATH,
        TEST_PATH,
        config_overrides={"language": "en"},
    )
    _ = pipeline.diagnose()
    training_artifact_info = pipeline.train()
    print("Selected method:", pipeline.training_method_cls.__name__)
    if training_artifact_info.test_metrics:
        print("Test F1:", training_artifact_info.test_metrics.f1)
    test_data = pipeline.splits.get("test")
    if test_data is not None and hasattr(pipeline.method_manager, "test"):
        training_artifact_info.test_metrics_inscope = pipeline.method_manager.test(
            test_data, inscope_only=True
        )
        if training_artifact_info.test_metrics_inscope:
            print(
                "Test F1 (in-scope only):",
                training_artifact_info.test_metrics_inscope.f1,
            )

    # Optional: save model and run inference
    # pipeline.save("saved_text_clf_torch", SaveFormat.TORCH)
    # pipeline.save("saved_text_clf_onnx", SaveFormat.ONNX)
    # manager = TextClassificationInferenceManager("saved_text_clf_torch")
    # outputs = manager.predict(["example sentence"], batch_size=32)
    # for out in outputs:
    #     print(out.most_probable.label, out.most_probable.score)


if __name__ == "__main__":
    main()



"""
Example: text classification with LLM-based augmentation for low-resource classes.
Requires MODEL_API_KEY and BASE_URL (and optionally MODEL_ID) in the environment.
"""

import os
import torch

from open_autonlu.auto_classes import TextClassificationTrainingPipeline

TRAIN_PATH = "examples/test_data/noise_n_shot_data/snips_200_300_0.01_42_train.csv"
TEST_PATH = "examples/test_data/noise_n_shot_data/snips_200_300_0.01_42_test.csv"


def main():
    api_key = os.environ.get("MODEL_API_KEY", "")
    base_url = os.environ.get("BASE_URL", "").strip()
    model_id = os.environ.get("MODEL_ID", "").strip()
    if not api_key or not base_url:
        print(
            "Set MODEL_API_KEY and BASE_URL (and optionally MODEL_ID) "
            "in the environment to run LLM augmentation."
        )
        return
    if not torch.cuda.is_available() and not torch.backends.mps.is_available():
        print("Warning: no GPU; training may be slow.")

    pipeline = TextClassificationTrainingPipeline(
        TRAIN_PATH,
        TEST_PATH,
        config_overrides={
            "language": "en",
            "llm_augmentation": {
                "enabled": True,
                "use_domain_analysis": True,
                "max_attempts": 20,
                "config_overrides": {
                    "LlmClientConfig": {
                        "api_key": api_key,
                        "base_url": base_url,
                        "model_id": model_id,
                    }
                },
            },
        },
    )
    _ = pipeline.diagnose()
    training_artifact_info = pipeline.train()
    print("Method:", pipeline.training_method_cls.__name__)
    if training_artifact_info.test_metrics:
        print("Test F1:", training_artifact_info.test_metrics.f1)


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Optional

from ..constants import DEFAULT_TEMPLATE_BY_LANGUAGE
from .setfit_config import SetFitMethodConfig


@dataclass
class AncSetFitConfig(SetFitMethodConfig):
    """Configuration for Anchor-based SetFit (AncSetFit) method.

    AncSetFit extends SetFit with anchor-based contrastive learning using triplet loss.
    Instead of comparing sentence pairs, it uses anchor templates that combine a prompt
    template with human-readable class descriptions (anc_label).

    The anc_label is a human-readable description of the class that gets concatenated
    with the template to form an anchor. For example, if template is
    "User asks the bot to perform a request using the skill: " and anc_label is
    "weather forecast", the anchor becomes "User asks the bot to perform a request
    using the skill: weather forecast". This anchor is then used in triplet loss
    training where:
    - Anchor: template + anc_label (class description)
    - Positive: sentence belonging to that class
    - Negative: sentence from a different class

    The anc_label column must be present in the training dataset. If absent,
    the pipeline will fall back to standard SetFit.

    Attributes:
        margin: Margin for triplet loss. Controls the minimum distance between
            positive and negative pairs.
        add_normalization_layer: Whether to add L2 normalization to embeddings.
        batch_size: Training batch size for triplet learning.
        epochs: Number of training epochs.
        body_lr: Learning rate for the sentence transformer body.
        template: Prompt template prepended to anc_label to form anchors.
            If None, resolved from language by DEFAULT_TEMPLATE_BY_LANGUAGE.
    """

    margin: float = 0.25
    add_normalization_layer: bool = False
    batch_size: int = 64
    epochs: int = 1
    body_lr: float = 1e-4
    template: Optional[str] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.template is None:
            self.template = DEFAULT_TEMPLATE_BY_LANGUAGE[self.language]



import logging
import os
from collections import Counter
from dataclasses import replace
from functools import partial
from typing import Any, Callable, Dict, List, Optional

import numpy as np
from accelerate.utils import find_executable_batch_size
from datasets import Dataset, DatasetDict, concatenate_datasets
from optimum.onnxruntime import ORTModelForSequenceClassification
from optimum.pipelines import pipeline as ort_pipeline
from optuna import Trial
from scipy.special import softmax
from sklearn.metrics import classification_report, confusion_matrix, f1_score
from sklearn.model_selection import train_test_split
from transformers import (
    AutoModelForSequenceClassification,
    AutoTokenizer,
    DataCollatorWithPadding,
    Trainer,
    set_seed,
)
from transformers.modeling_utils import PreTrainedModel
from transformers.trainer_callback import EarlyStoppingCallback

from ..constants import (
    DATA_PROCESSING_STAGE,
    DEV_EVALUATION_STAGE,
    HPO_STAGE,
    TRAINING_STAGE,
)
from ..methods.data_types import OodMethod
from ..methods.constants import OOS_LABEL
from ..methods.utils.misc import initialize_transformers_tokenizer
from ..progress_reporter import (
    ProgressReporter,
    TrainerProgressReporterCallback,
)
from .configs import FinetunerConfig
from .constants import BATCH_SIZE
from .data_types import (
    ClassificationEvaluationResult,
    SaveFormat,
    SimplePyTorchDataset,
    TextClassifierOutput,
    TextLabel,
    TrainingArtifactInfo,
)
from .hyp_search_mixin import FinetuningHyperparameterSearchMixin
from .method import SequenceClassificationMethod
from .onnx_inference_manager import AbstractONNXInferenceManager
from .utils import GibberishDatasetGenerator

log = logging.getLogger(__name__)


class FinetunerBase(SequenceClassificationMethod, FinetuningHyperparameterSearchMixin):
    """Base class for finetuning sequence classification models.

    Finetunes a pretrained transformer model for text classification.
    Supports hyperparameter optimization via Optuna and early stopping.

    Attributes:
        config_cls: Configuration class for this method.
        stages: List of training stages for progress reporting.
    """

    config_cls = FinetunerConfig
    stages = [DATA_PROCESSING_STAGE, HPO_STAGE, TRAINING_STAGE, DEV_EVALUATION_STAGE]

    def __init__(
        self,
        config: Optional[FinetunerConfig] = None,
        progress_reporter: Optional[ProgressReporter] = None,
    ):
        """Initialize the Finetuner.

        Args:
            config: Configuration object with model and training parameters.
                If None, uses default FinetunerConfig.
            progress_reporter: Reporter for tracking training progress.
                If None, creates a default reporter.
        """
        if config is None:
            config = FinetunerConfig()
        if progress_reporter is None:
            progress_reporter = ProgressReporter()
        self.progress_reporter = progress_reporter
        self.model_name = config.model_name_or_path
        self.num_hpo_trials = config.num_hpo_trials
        self.training_args = config.training_arguments
        self.seed = self.training_args.seed
        self.ood_method = config.ood_method

        self.training_args = replace(
            self.training_args,
            report_to="none",
            overwrite_output_dir=True,
        )
        self.dev_fraction = config.dev_fraction
        self.tokenizer = initialize_transformers_tokenizer(
            self.model_name, config.max_seq_length, config.tokenizer_kwargs
        )
        self.trainer = None
        self.model = None
        self.id2label = None
        self.label2id = None
        self.hyp_search_done = False
        self.progress_reporter = progress_reporter

    @property
    def id2label(self):
        if getattr(self, "model"):
            return self.model.config.id2label
        else:
            return self._id2label

    @id2label.setter
    def id2label(self, value):
        self._id2label = value

    def preprocess_function(self, examples):
        """Takes a batch of examples and tokenizes it. Adds labels to the batch if present"""
        tokenized_examples = self.tokenizer(examples["text"], truncation=True)
        labels = examples.get("label")
        if labels is not None:
            if self.model:
                label2id = self.model.config.label2id
            else:
                label2id = self.label2id
            tokenized_examples["labels"] = [label2id[label] for label in labels]
        return tokenized_examples

    @staticmethod
    def _postprocess_logits(logits: np.ndarray) -> np.ndarray:
        return softmax(logits, axis=-1)

    def compute_metrics_train(self, eval_preds):
        """Computes macro f1 metric during training"""
        logits, labels = eval_preds
        if isinstance(logits, tuple):
            logits = logits[0]
        predictions = np.argmax(logits, axis=-1)
        return {"f1": f1_score(labels, predictions, average="macro")}

    def compute_metrics_eval(self, eval_preds, id2label=None):
        logits, labels = eval_preds
        labels = [id2label[label] for label in labels]
        if isinstance(logits, tuple):
            logits = logits[0]
        labels_hat = self._extract_labels(logits, id2label)
        clf_report = classification_report(
            labels,
            labels_hat,
            output_dict=True,
        )
        result = {
            "classification_report": clf_report,
            "confusion_matrix": confusion_matrix(
                y_true=labels, y_pred=labels_hat, labels=self.label_list
            ),
            "f1": clf_report["macro avg"]["f1-score"],
        }
        return result

    def _split_data(self, train_dataset: Dataset) -> DatasetDict:
        train_ids, eval_ids = train_test_split(
            range(train_dataset.num_rows),
            test_size=self.dev_fraction,
            random_state=self.seed,
            stratify=train_dataset["label"],
        )
        return DatasetDict(
            {
                "train": train_dataset.select(train_ids),
                "eval": train_dataset.select(eval_ids),
            }
        )

    def _generate_synthetic_oos(self, num_train_rows: int) -> Dataset:
        generator = GibberishDatasetGenerator(self.training_args.seed)
        synthetic_texts = generator(num_rows=num_train_rows)
        synthetic_dataset = Dataset.from_dict(
            {"text": synthetic_texts, "label": [OOS_LABEL] * num_train_rows}
        )
        return synthetic_dataset

    def train(self, train_dataset: Dataset) -> TrainingArtifactInfo:
        """
        Trains the model using the provided training dataset.

        Implementation Details:
            - The method first splits the training dataset into training and evaluation subsets.
            - It then initializes the model, data collator, and trainer.
            - If hyperparameter optimization (HPO) is enabled, it performs HPO using the specified search space and number of trials.
            - After HPO, it reconfigures the trainer with the best hyperparameters found.
            - Finally, it trains the model using the trainer and updates the model attribute with the trained model.

        Args:
            train_dataset (Dataset): The training dataset containing the labeled data.

        Returns:
            TrainingArtifactInfo: Information about the trained model and its performance.
        """
        self.progress_reporter.stages = self.stages
        self.progress_reporter(DATA_PROCESSING_STAGE, 0.0)
        set_seed(self.seed)

        label_counts = Counter(train_dataset["label"])
        ood_count = label_counts.get(OOS_LABEL, 0)
        p95 = int(np.percentile(list(label_counts.values()), 95))
        if self.ood_method == OodMethod.NONE:
            if ood_count > 0:
                synthetic_texts_ds = self._generate_synthetic_oos(p95 - ood_count)
                train_dataset = concatenate_datasets(
                    [train_dataset, synthetic_texts_ds]
                )
        elif self.ood_method == OodMethod.LOGIT:
            if ood_count == 0:
                synthetic_texts_ds = self._generate_synthetic_oos(p95)
                train_dataset = concatenate_datasets(
                    [train_dataset, synthetic_texts_ds]
                )
            elif ood_count < p95:
                synthetic_texts_ds = self._generate_synthetic_oos(p95 - ood_count)
                train_dataset = concatenate_datasets(
                    [train_dataset, synthetic_texts_ds]
                )

        labels = train_dataset.unique("label")
        self.num_labels = len(labels)
        data_collator = DataCollatorWithPadding(tokenizer=self.tokenizer)
        self.label2id, self.id2label = self._get_label_mappings(labels)
        data = self._split_data(train_dataset).map(
            self.preprocess_function,
            remove_columns=train_dataset.column_names,
            batched=True,
        )
        self.progress_reporter(DATA_PROCESSING_STAGE, 1.0)
        self.training_args.configure_for_early_stopping()
        self.trainer = Trainer(
            model_init=self._prepare_model_init_function(),
            args=self.training_args,
            train_dataset=data["train"],
            eval_dataset=data["eval"],
            data_collator=data_collator,
            tokenizer=self.tokenizer,
            compute_metrics=self.compute_metrics_train,
            callbacks=[EarlyStoppingCallback(early_stopping_patience=5)],
        )
        if self.training_args.hyp_search_space and not self.hyp_search_done:
            log.info("Running hyperparameter optimization")
            self.progress_reporter(HPO_STAGE, 0.0)
            best_params = self.optimize_hyperparameters(
                self.trainer,
                self.training_args.hyp_search_space,
                self.num_hpo_trials,
                progress_reporter=self.progress_reporter,
            )
            self.progress_reporter(HPO_STAGE, 1.0)
            self.training_args = replace(self.training_args, **best_params)
            self.training_args.configure_for_early_stopping()
            self.trainer = Trainer(
                model_init=self._prepare_model_init_function(),
                args=self.training_args,
                train_dataset=data["train"],
                eval_dataset=data["eval"],
                data_collator=data_collator,
                tokenizer=self.tokenizer,
                compute_metrics=self.compute_metrics_train,
                callbacks=[
                    EarlyStoppingCallback(early_stopping_patience=5),
                    TrainerProgressReporterCallback(
                        self.progress_reporter, TRAINING_STAGE
                    ),
                ],
            )

            self.hyp_search_done = True
        self.trainer.train()
        self.model = self.trainer.model
        self.trainer.compute_metrics = partial(
            self.compute_metrics_eval, id2label=self.id2label
        )
        self.progress_reporter(DEV_EVALUATION_STAGE, 0.0)
        metrics: Dict[str, Any] = self.trainer.evaluate(metric_key_prefix="")
        self.progress_reporter(DEV_EVALUATION_STAGE, 1.0)

        # Set actual train batch size before reporting
        if (
            hasattr(self.training_args, "auto_find_batch_size")
            and self.training_args.auto_find_batch_size is True
        ):
            self.training_args.per_device_train_batch_size = (
                self.trainer.state.train_batch_size
            )

        return TrainingArtifactInfo(
            self.training_args.to_dict(),
            self.label_list,
            ClassificationEvaluationResult(
                classification_report=metrics["_classification_report"],
                confusion_matrix=metrics["_confusion_matrix"],
                f1=metrics["_f1"],
            ),
        )

    def test(
        self, dataset: Dataset, inscope_only: bool = False
    ) -> ClassificationEvaluationResult:
        """Evaluate the model on a test dataset.

        Args:
            dataset: Test dataset with 'text' and 'label' columns.
            inscope_only: If True, exclude OOS predictions and labels from
                evaluation, replacing OOS predictions with the most probable
                in-scope label.

        Returns:
            ClassificationEvaluationResult with confusion matrix,
            classification report, and macro F1 score.
        """
        # Mitigating the issue with missing labels in model label set
        # by removing labels from the dataset before preprocessing
        # so that preprocess_finction will not try to map unknown labels to ids
        columns_to_remove = dataset.column_names
        columns_to_remove.remove("label")
        preprocessed_inputs = dataset.remove_columns("label").map(
            self.preprocess_function, remove_columns=columns_to_remove, batched=True
        )
        prediction_output = self.trainer.predict(preprocessed_inputs).predictions
        labels_hat = self._extract_labels(prediction_output, self.id2label)

        if inscope_only:
            oos_label_id = self.label2id.get(OOS_LABEL)
            if oos_label_id is not None:
                inscope_label_ids = [
                    label_id
                    for label_id, label in self.id2label.items()
                    if label != OOS_LABEL
                ]
                if inscope_label_ids:
                    for i, pred_label in enumerate(labels_hat):
                        if pred_label == OOS_LABEL:
                            logits = prediction_output[i]
                            masked_logits = logits.copy()
                            masked_logits[oos_label_id] = -np.inf
                            top_inscope_label_id = inscope_label_ids[
                                np.argmax(
                                    [
                                        masked_logits[label_id]
                                        for label_id in inscope_label_ids
                                    ]
                                )
                            ]
                            labels_hat[i] = self.id2label[top_inscope_label_id]

            original_labels = dataset["label"]
            mask = [label != OOS_LABEL for label in original_labels]
            dataset = dataset.filter(lambda example, idx: mask[idx], with_indices=True)
            labels_hat = [label for label, keep in zip(labels_hat, mask) if keep]

        ds_label_set = set(dataset.unique("label"))
        if inscope_only and OOS_LABEL in ds_label_set:
            ds_label_set.remove(OOS_LABEL)

        model_label_set = set(self.label_list)
        extra_in_test = ds_label_set - model_label_set
        if extra_in_test:
            log.warning(
                "Test dataset contains labels not seen during training: %s",
                sorted(extra_in_test),
            )

        clf_report = classification_report(
            dataset["label"],
            labels_hat,
            labels=sorted(ds_label_set),
            output_dict=True,
            zero_division=0,
        )
        # Mitigating the issue with missing labels in model label set
        # while retaining the order of labels in the confusion matrix
        labels_for_conf_matrix = self.label_list
        if ds_label_set != model_label_set:
            labels_for_conf_matrix = self.label_list + list(
                ds_label_set - model_label_set
            )

        conf_matrix = confusion_matrix(
            dataset["label"], labels_hat, labels=labels_for_conf_matrix
        )
        return ClassificationEvaluationResult(
            conf_matrix, clf_report, clf_report["macro avg"]["f1-score"]
        )

    def _predict(
        self,
        batch_size: int,
        texts: List[str],
    ) -> list[TextClassifierOutput]:
        if isinstance(texts, str):
            texts = [texts]
        preprocessed_inputs = SimplePyTorchDataset(
            self.tokenizer(texts, truncation=True).data
        )
        if batch_size != self.training_args.per_device_eval_batch_size:
            self.training_args.per_device_eval_batch_size = batch_size
        prediction_output = self.trainer.predict(preprocessed_inputs).predictions
        prediction_output = self._postprocess_logits(prediction_output)
        labels = self._extract_labels(prediction_output, self.id2label)
        scores = self._extract_scores(prediction_output)
        output = [
            TextClassifierOutput(text=text, hypotheses=[TextLabel(label, score)])
            for label, score, text in zip(labels, scores, texts)
        ]
        return output

    def save(
        self, path_to_package: str, checkpoint_format: SaveFormat = SaveFormat.TORCH
    ) -> None:
        """Save the trained model to disk.

        Args:
            path_to_package: Directory path for saving the model.
            checkpoint_format: Format to save in. TORCH saves the full
                model; ONNX exports an optimized inference model.
        """
        self.trainer.save_model(path_to_package)
        if checkpoint_format == SaveFormat.ONNX:
            ort_model = ORTModelForSequenceClassification.from_pretrained(
                path_to_package, export=True
            )
            ort_model.save_pretrained(path_to_package)
            os.unlink(path_to_package + "/model.safetensors")
        log.info("Model is saved at " + path_to_package)

    @classmethod
    def load(cls, path_to_package: str):
        """Load a trained Finetuner model from disk.

        Args:
            path_to_package: Directory path containing the saved model.

        Returns:
            Initialized FinetunerBase instance with loaded model.
        """
        self = cls()
        self.model = AutoModelForSequenceClassification.from_pretrained(
            path_to_package, return_dict=True
        )
        self.trainer = Trainer(
            self.model,
            tokenizer=self.tokenizer,
            data_collator=DataCollatorWithPadding(self.tokenizer),
            args=self.training_args,
        )
        log.info("Model is loaded and ready to use")
        return self

    def _prepare_model_init_function(
        self, **model_kwargs
    ) -> Callable[[Trial], PreTrainedModel]:
        default_model_kwargs = {
            "output_attentions": False,
            "output_hidden_states": False,
            "return_dict": True,
        }
        if not model_kwargs:
            model_kwargs = default_model_kwargs
        else:
            default_model_kwargs.update(model_kwargs)
            model_kwargs = default_model_kwargs

        def model_init(trial: Trial) -> PreTrainedModel:
            return AutoModelForSequenceClassification.from_pretrained(
                self.model_name,
                num_labels=self.num_labels,
                id2label=self.id2label,
                label2id=self.label2id,
                **model_kwargs,
            )

        return model_init


class FinetunerInferenceManager(AbstractONNXInferenceManager):
    """ONNX-based inference manager for finetuned models.

    Uses ONNX Runtime for optimized inference on finetuned
    sequence classification models.
    """

    def __init__(self, path_to_package: str) -> None:
        """Initialize the inference manager.

        Args:
            path_to_package: Path to directory with saved ONNX model.
        """
        super().__init__(path_to_package)
        self.model = ORTModelForSequenceClassification.from_pretrained(
            path_to_package, provider=self.ort_provider
        )
        self.tokenizer = AutoTokenizer.from_pretrained(path_to_package)
        self._pipeline = ort_pipeline(
            "text-classification",
            model=self.model,
            tokenizer=self.tokenizer,  # type: ignore
            accelerator="ort",
        )

    def predict(
        self,
        texts: List[str],
        batch_size: int = BATCH_SIZE,
        auto_find_batch_size: bool = False,
        return_all_hypotheses: bool = False,
    ) -> list[TextClassifierOutput]:
        """Run inference on a list of texts.

        Args:
            texts: List of text strings to classify.
            batch_size: Number of texts to process in each batch.
            auto_find_batch_size: If True, automatically reduce batch size
                on OOM errors.
            return_all_hypotheses: Not supported for ONNX inference.

        Returns:
            List of TextClassifierOutput with predicted labels and scores.
        """
        if return_all_hypotheses:
            log.debug(
                "return_all_hypotheses is not supported for ONNX-based inference and will be ignored."
            )
        if auto_find_batch_size:
            return find_executable_batch_size(
                self._predict, starting_batch_size=batch_size
            )(texts)  # type: ignore
        return self._predict(batch_size, texts)

    def _predict(self, batch_size: int, texts: List[str]) -> list[TextClassifierOutput]:
        return [
            TextClassifierOutput(
                text=text, hypotheses=[TextLabel(result["label"], result["score"])]
            )
            for result, text in zip(
                self._pipeline(
                    texts, batch_size=batch_size, truncation=True, padding=True
                ),
                texts,
            )  # type: ignore
        ]


class Finetuner(FinetunerBase):
    """Finetuner method for text classification.

    Concrete implementation of FinetunerBase with ONNX inference support.
    This is the recommended finetuning variant for production use.
    """

    onnx_inference_manager = FinetunerInferenceManager



from .misc import batch, resolve_device
from .ner_metrics import evaluate_entity_level
from .generic_sequence_classifier_onnx_exporter import (
    GenericSequenceClassifierONNXExportMixin,
)
from .random_sentence_generator import (
    GibberishDatasetGenerator,
    GibberishDatasetGeneratorEN,
)
from .sentence_transformers_fit import fit_encoder

__all__ = [
    "batch",
    "evaluate_entity_level",
    "GenericSequenceClassifierONNXExportMixin",
    "GibberishDatasetGenerator",
    "GibberishDatasetGeneratorEN",
    "fit_encoder",
    "resolve_device",
]



from .text_classification_pipeline import (
    TextClassificationTrainingPipeline,
    TextClassificationInferenceManager,
)
from .token_classification_pipeline import (
    TokenClassificationTrainingPipeline,
    TokenClassificationInferenceManager,
)

__all__ = [
    "TextClassificationTrainingPipeline",
    "TextClassificationInferenceManager",
    "TokenClassificationTrainingPipeline",
    "TokenClassificationInferenceManager",
]



import re
from augmentex import WordAug, CharAug
import random
import numpy as np


class WordCharAugmentation:
    """Text augmentation class that applies character-level and word-level corruptions.

    Attributes:
        char_actions: List of character-level augmentation actions.
        word_actions: List of word-level augmentation actions.
        seed: Random seed for reproducibility (default: 42).
        word_aug: WordAug instance from augmentex.
        char_aug: CharAug instance from augmentex.

    Example:
        >>> aug = WordCharAugmentation()
        >>> corruptions = aug("Привет мир")
        >>> print(corruptions)
        ['Привет мри', 'Прривет мир', 'Првет мир', ...]
    """

    def __init__(self, *args, **kwargs):
        self.char_actions = ["orfo", "typo", "delete", "multiply", "swap", "insert"]
        self.word_actions = ["replace", "swap"]
        self.seed = 42
        self.word_aug = WordAug(
            unit_prob=1.0,
            # Percentage of the phrase to which augmentations will be applied
            min_aug=1,  # Minimum number of augmentations
            max_aug=2,  # Maximum number of augmentations
            # following arguments are supported in augmentex 1.3.1
            lang="rus",  # supports: "rus", "eng"
            platform="pc",  # supports: "pc", "mobile"
            random_seed=self.seed,
        )
        self.char_aug = CharAug(
            unit_prob=1.0,  # Percentage of the phrase to which augmentations will be applied
            min_aug=1,  # Minimum number of augmentations
            max_aug=2,  # Maximum number of augmentations
            lang="rus",  # supports: "rus", "eng"
            platform="pc",  # supports: "pc", "mobile"
            random_seed=self.seed,
        )
        self._set_seed(self.seed)

    @staticmethod
    def _clean_for_aug(x: str) -> str:
        return re.sub("[^\\sа-яА-ЯёЁa-zA-Z0-9'.,!?;\\]\\[]", "", x.replace("-", " "))

    def _set_seed(self, seed: int = None):
        if seed is None:
            seed = 42
        random.seed(seed)
        np.random.seed(seed)

    def __call__(self, text: str) -> list[str]:
        corruptions = []
        for action in self.char_actions:
            try:
                self._set_seed()
                aug_text = self.char_aug.augment(text, action=action)
                corruptions.append(aug_text)
            except Exception as e:
                print(f"Augmentation '{action}' failed on text '{text}': {e}")
        for action in self.word_actions:
            try:
                cleaned_text = self._clean_for_aug(text)
                self._set_seed()
                aug_text = self.word_aug.augment(cleaned_text, action=action)
                corruptions.append(aug_text)
            except Exception as e:
                print(f"Augmentation '{action}' failed on text '{text}': {e}")

        corruptions = sorted(list(set(corruptions)))
        return corruptions



import enum
from dataclasses import dataclass
from typing import Iterable, Optional
from datasets import DatasetDict


class TaskType(enum.Enum):
    MULTICLASS = "multiclass_classification"
    MULTILABEL = "multilabel_classification"
    NER = "token_classification"


@dataclass
class DatasetEvaluatorOutput:
    splits_filtered: DatasetDict
    indices_to_remove: Iterable[int]
    aggregated_report_path: Optional[str]
    reports_by_method_folder: Optional[str]



import logging
from pathlib import Path
from typing import Callable, Iterable, Optional

import torch
import transformers
from datasets import Dataset, DatasetDict
from packaging import version
from sentence_transformers import SentenceTransformer
from sentence_transformers.datasets import NoDuplicatesDataLoader, SentenceLabelDataset
from sentence_transformers.evaluation import SentenceEvaluator
from sentence_transformers.fit_mixin import (
    EvaluatorCallback,
    OriginalCallback,
    SaveModelCallback,
)
from sentence_transformers.training_args import (
    BatchSamplers,
    MultiDatasetBatchSamplers,
    SentenceTransformerTrainingArguments,
)
from sentence_transformers.util import is_datasets_available
from torch import nn
from torch.optim import Optimizer
from torch.utils.data import DataLoader

from ...constants import PRETRAINING_STAGE
from ...progress_reporter import (
    ProgressReporter,
    TrainerProgressReporterCallback,
)

logger = logging.getLogger(__name__)


# re-implementation of fit mixin from https://github.com/UKPLab/sentence-transformers/blob/7d52a069e0b37d976b3ed3f674a6180436c27574/sentence_transformers/fit_mixin.py#L162
# for better control
def fit_encoder(
    encoder: SentenceTransformer,
    train_objectives: Iterable[tuple[DataLoader, nn.Module]],
    evaluator: SentenceEvaluator = None,
    epochs: int = 1,
    steps_per_epoch=None,
    scheduler: str = "WarmupLinear",
    warmup_steps: int = 10000,
    optimizer_class: type[Optimizer] = torch.optim.AdamW,
    optimizer_params: dict[str, object] = {"lr": 2e-5},
    weight_decay: float = 0.01,
    evaluation_steps: int = 0,
    output_path: str = None,
    save_best_model: bool = True,
    max_grad_norm: float = 1,
    use_amp: bool = False,
    callback: Callable[[float, int, int], None] = None,
    show_progress_bar: bool = True,
    checkpoint_path: str = None,
    checkpoint_save_steps: int = 500,
    checkpoint_save_total_limit: int = 0,
    progress_reporter: Optional[ProgressReporter] = None,
) -> SentenceTransformer:
    """
    Deprecated training method from before Sentence Transformers v3.0, it is recommended to use
    :class:`~sentence_transformers.trainer.SentenceTransformerTrainer` instead. This method uses
    :class:`~sentence_transformers.trainer.SentenceTransformerTrainer` behind the scenes, but does
    not provide as much flexibility as the Trainer itself.

    This training approach uses a list of DataLoaders and Loss functions to train the model. Each DataLoader
    is sampled in turn for one batch. We sample only as many batches from each DataLoader as there are in the
    smallest one to make sure of equal training with each dataset, i.e. round robin sampling.

    This method should produce equivalent results in v3.0+ as before v3.0, but if you encounter any issues
    with your existing training scripts, then you may wish to use
    :meth:`SentenceTransformer.old_fit <sentence_transformers.SentenceTransformer.old_fit>` instead.
    That uses the old training method from before v3.0.

    Args:
        encoder: SentenceTransformer model to train.
        train_objectives: Tuples of (DataLoader, LossFunction). Pass
            more than one for multi-task learning
        evaluator: An evaluator (sentence_transformers.evaluation)
            evaluates the model performance during training on held-
            out dev data. It is used to determine the best model
            that is saved to disc.
        epochs: Number of epochs for training
        steps_per_epoch: Number of training steps per epoch. If set
            to None (default), one epoch is equal the DataLoader
            size from train_objectives.
        scheduler: Learning rate scheduler. Available schedulers:
            constantlr, warmupconstant, warmuplinear, warmupcosine,
            warmupcosinewithhardrestarts
        warmup_steps: Behavior depends on the scheduler. For
            WarmupLinear (default), the learning rate is increased
            from o up to the maximal learning rate. After these many
            training steps, the learning rate is decreased linearly
            back to zero.
        optimizer_class: Optimizer
        optimizer_params: Optimizer parameters
        weight_decay: Weight decay for model parameters
        evaluation_steps: If > 0, evaluate the model using evaluator
            after each number of training steps
        output_path: Storage path for the model and evaluation files
        save_best_model: If true, the best model (according to
            evaluator) is stored at output_path
        max_grad_norm: Used for gradient normalization.
        use_amp: Use Automatic Mixed Precision (AMP). Only for
            Pytorch >= 1.6.0
        callback: Callback function that is invoked after each
            evaluation. It must accept the following three
            parameters in this order: `score`, `epoch`, `steps`
        show_progress_bar: If True, output a tqdm progress bar
        checkpoint_path: Folder to save checkpoints during training
        checkpoint_save_steps: Will save a checkpoint after so many
            steps
        checkpoint_save_total_limit: Total number of checkpoints to
            store
    """
    if not is_datasets_available():
        raise ImportError(
            "Please install `datasets` to use this function: `pip install datasets`."
        )

    # Delayed import to counter the SentenceTransformers -> FitMixin -> SentenceTransformerTrainer -> SentenceTransformers circular import
    from sentence_transformers.trainer import SentenceTransformerTrainer

    data_loaders, loss_fns = zip(*train_objectives)

    # Clear the dataloaders from collate functions as we just want raw InputExamples
    def identity(batch):
        return batch

    for data_loader in data_loaders:
        data_loader.collate_fn = identity

    batch_size = 8
    batch_sampler = BatchSamplers.BATCH_SAMPLER
    # Convert dataloaders into a DatasetDict
    # TODO: This is rather inefficient, as we load all data into memory. We might benefit from a more efficient solution
    train_dataset_dict = {}
    for loader_idx, data_loader in enumerate(data_loaders, start=1):
        if isinstance(data_loader, NoDuplicatesDataLoader):
            batch_sampler = BatchSamplers.NO_DUPLICATES
        elif hasattr(data_loader, "dataset") and isinstance(
            data_loader.dataset, SentenceLabelDataset
        ):
            batch_sampler = BatchSamplers.GROUP_BY_LABEL

        batch_size = getattr(data_loader, "batch_size", batch_size)
        texts = []
        labels = []
        for batch in data_loader:
            batch_texts, batch_labels = zip(
                *[(example.texts, example.label) for example in batch]
            )
            texts += batch_texts
            labels += batch_labels
        dataset = Dataset.from_dict(
            {f"sentence_{idx}": text for idx, text in enumerate(zip(*texts))}
        )
        # Add label column, unless all labels are 0 (the default value for `labels` in InputExample)
        add_label_column = True
        try:
            if set(labels) == {0}:
                add_label_column = False
        except TypeError:
            pass
        if add_label_column:
            dataset = dataset.add_column("label", labels)
        train_dataset_dict[f"_dataset_{loader_idx}"] = dataset

    train_dataset_dict = DatasetDict(train_dataset_dict)

    def _default_checkpoint_dir() -> str:
        dir_name = "checkpoints/model"
        idx = 1
        while Path(dir_name).exists() and len(list(Path(dir_name).iterdir())) != 0:
            dir_name = f"checkpoints/model_{idx}"
            idx += 1
        return dir_name

    # Convert loss_fns into a dict with `dataset_{idx}` keys
    loss_fn_dict = {
        f"_dataset_{idx}": loss_fn for idx, loss_fn in enumerate(loss_fns, start=1)
    }

    # Use steps_per_epoch to perhaps set max_steps
    max_steps = -1
    if steps_per_epoch is not None and steps_per_epoch > 0:
        if epochs == 1:
            max_steps = steps_per_epoch
        else:
            logger.warning(
                "Setting `steps_per_epoch` alongside `epochs` > 1 no longer works. "
                "We will train with the full datasets per epoch."
            )
            steps_per_epoch = None

    # Transformers renamed `evaluation_strategy` to `eval_strategy` in v4.41.0
    eval_strategy_key = (
        "eval_strategy"
        if version.parse(transformers.__version__) >= version.parse("4.41.0")
        else "evaluation_strategy"
    )
    args = SentenceTransformerTrainingArguments(
        output_dir=checkpoint_path or _default_checkpoint_dir(),
        batch_sampler=batch_sampler,
        multi_dataset_batch_sampler=MultiDatasetBatchSamplers.ROUND_ROBIN,
        per_device_train_batch_size=batch_size,
        per_device_eval_batch_size=batch_size,
        num_train_epochs=epochs,
        max_steps=max_steps,
        **{
            eval_strategy_key: "steps"
            if evaluation_steps is not None and evaluation_steps > 0
            else "no",
        },
        eval_steps=evaluation_steps,
        # load_best_model_at_end=save_best_model, # <- TODO: Look into a good solution for save_best_model
        max_grad_norm=max_grad_norm,
        fp16=use_amp,
        disable_tqdm=not show_progress_bar,
        save_strategy="steps" if checkpoint_path is not None else "no",
        save_steps=checkpoint_save_steps,
        save_total_limit=checkpoint_save_total_limit,
        report_to=["none"],
    )

    if steps_per_epoch is None or steps_per_epoch == 0:
        steps_per_epoch = min(
            [
                len(train_dataset) // batch_size
                for train_dataset in train_dataset_dict.values()
            ]
        )
    num_train_steps = int(steps_per_epoch * epochs)

    # Prepare optimizer & scheduler
    param_optimizer = list(encoder.named_parameters())

    no_decay = ["bias", "LayerNorm.bias", "LayerNorm.weight"]
    optimizer_grouped_parameters = [
        {
            "params": [
                p for n, p in param_optimizer if not any(nd in n for nd in no_decay)
            ],
            "weight_decay": weight_decay,
        },
        {
            "params": [
                p for n, p in param_optimizer if any(nd in n for nd in no_decay)
            ],
            "weight_decay": 0.0,
        },
    ]

    optimizer = optimizer_class(optimizer_grouped_parameters, **optimizer_params)
    scheduler_obj = encoder._get_scheduler(
        optimizer,
        scheduler=scheduler,
        warmup_steps=warmup_steps,
        t_total=num_train_steps,
    )

    # Create callbacks
    callbacks = []
    if evaluator is not None:
        callbacks.append(EvaluatorCallback(evaluator, output_path))
        if callback is not None:
            callbacks.append(OriginalCallback(callback, evaluator))
    if progress_reporter is not None:
        callbacks.append(
            TrainerProgressReporterCallback(progress_reporter, PRETRAINING_STAGE)
        )

    trainer = SentenceTransformerTrainer(
        model=encoder,
        args=args,
        train_dataset=train_dataset_dict,
        eval_dataset=None,
        loss=loss_fn_dict,
        evaluator=evaluator,
        optimizers=(optimizer, scheduler_obj),
        callbacks=callbacks,
    )
    # Set the trainer on the EvaluatorCallback, required for logging the metrics
    for callback in trainer.callback_handler.callbacks:
        if isinstance(callback, EvaluatorCallback):
            callback.trainer = trainer

    if output_path is not None:
        trainer.add_callback(SaveModelCallback(output_path, evaluator, save_best_model))

    trainer.train()
    return trainer.model



import logging
from copy import deepcopy
from typing import List, Dict, Union, Tuple, Optional

import numpy as np

from .utils import conll_to_spans, find_overlap, list_to_spans

logger = logging.getLogger(__name__)


class Evaluator:  # pylint: disable=too-many-instance-attributes, too-few-public-methods
    def __init__(
        self,
        true: Union[
            List[List[str]],
            List[str],
            List[Dict],
            str,
            List[List[Dict[str, Union[int, str]]]],
        ],
        pred: Union[
            List[List[str]],
            List[str],
            List[Dict],
            str,
            List[List[Dict[str, Union[int, str]]]],
        ],
        tags: List[str],
        loader: str = "default",
    ) -> None:
        self.true = true
        self.pred = pred
        self.tags = tags
        # self.list = []

        # Setup dict into which metrics will be stored.
        self.metrics_results = {
            "correct": 0,
            "incorrect": 0,
            "partial": 0,
            "missed": 0,
            "spurious": 0,
            "possible": 0,
            "actual": 0,
            "precision": 0,
            "recall": 0,
            "f1": 0,
        }

        # Copy results dict to cover the four schemes.
        self.results = {
            "strict": deepcopy(self.metrics_results),
            "ent_type": deepcopy(self.metrics_results),
            "partial": deepcopy(self.metrics_results),
            "exact": deepcopy(self.metrics_results),
        }

        # Create an accumulator to store results
        self.evaluation_agg_entities_type = {e: deepcopy(self.results) for e in tags}
        self.loaders = {
            "list": list_to_spans,
            "conll": conll_to_spans,
        }

        self.loader = loader

        self.eval_indices: Dict[str, List[int]] = {
            "correct_indices": [],
            "incorrect_indices": [],
            "partial_indices": [],
            "missed_indices": [],
            "spurious_indices": [],
        }

        # Create dicts to hold indices for correct/spurious/missing/etc examples
        self.evaluation_indices = {
            "strict": deepcopy(self.eval_indices),
            "ent_type": deepcopy(self.eval_indices),
            "partial": deepcopy(self.eval_indices),
            "exact": deepcopy(self.eval_indices),
        }
        self.evaluation_agg_indices = {
            e: deepcopy(self.evaluation_indices) for e in tags
        }

    def evaluate(self) -> Tuple[Dict, Dict, Dict, Dict, np.ndarray]:
        logging.debug(
            "Imported %s predictions for %s true examples",
            len(self.pred),
            len(self.true),
        )

        if self.loader != "default":
            loader = self.loaders[self.loader]
            self.pred = loader(self.pred)
            self.true = loader(self.true)

        if len(self.true) != len(self.pred):
            raise ValueError("Number of predicted documents does not equal true")

        confusion_matrix = np.zeros((len(self.tags), len(self.tags)))

        for index, (true_ents, pred_ents) in enumerate(zip(self.true, self.pred)):
            # Compute results for one message
            (
                tmp_results,
                tmp_agg_results,
                tmp_results_indices,
                tmp_agg_results_indices,
                tmp_confusion_matrix,
            ) = compute_metrics(true_ents, pred_ents, self.tags, index)

            confusion_matrix += tmp_confusion_matrix

            # Cycle through each result and accumulate
            # TODO: Combine these loops below:
            for eval_schema in self.results:
                for metric in self.results[eval_schema]:
                    self.results[eval_schema][metric] += tmp_results[eval_schema][
                        metric
                    ]

                # Accumulate indices for each error type
                for error_type in self.evaluation_indices[eval_schema]:
                    self.evaluation_indices[eval_schema][error_type] += (
                        tmp_results_indices[eval_schema][error_type]
                    )

            # Calculate global precision and recall
            self.results = compute_precision_recall_wrapper(self.results)

            # Aggregate results by entity type
            for label in self.tags:
                for eval_schema in tmp_agg_results[label]:
                    for metric in tmp_agg_results[label][eval_schema]:
                        self.evaluation_agg_entities_type[label][eval_schema][
                            metric
                        ] += tmp_agg_results[label][eval_schema][metric]

                    # Accumulate indices for each error type per entity type
                    for error_type in self.evaluation_agg_indices[label][eval_schema]:
                        self.evaluation_agg_indices[label][eval_schema][error_type] += (
                            tmp_agg_results_indices[label][eval_schema][error_type]
                        )

                # Calculate precision recall at the individual entity level
                self.evaluation_agg_entities_type[label] = (
                    compute_precision_recall_wrapper(
                        self.evaluation_agg_entities_type[label]
                    )
                )

        return (
            self.results,
            self.evaluation_agg_entities_type,
            self.evaluation_indices,
            self.evaluation_agg_indices,
            confusion_matrix,
        )


# flake8: noqa: C901
def compute_metrics(  # type: ignore
    true_named_entities, pred_named_entities, tags: List[str], instance_index: int = 0
):  # pylint: disable=too-many-locals, too-many-branches, too-many-statements
    """
    Compute metrics on the collected true and predicted named entities

    :true_name_entities:
        collected true named entities output by collect_named_entities

    :pred_name_entities:
        collected predicted named entities output by collect_named_entities

    :tags:
        list of tags to be used

    :instance_index:
        index of the example being evaluated. Used to record indices of correct/missing/spurious/exact/partial predictions.
    """

    eval_metrics = {
        "correct": 0,
        "incorrect": 0,
        "partial": 0,
        "missed": 0,
        "spurious": 0,
        "precision": 0,
        "recall": 0,
        "f1": 0,
    }

    # overall results
    evaluation = {
        "strict": deepcopy(eval_metrics),
        "ent_type": deepcopy(eval_metrics),
        "partial": deepcopy(eval_metrics),
        "exact": deepcopy(eval_metrics),
    }

    # results by entity type
    evaluation_agg_entities_type = {e: deepcopy(evaluation) for e in tags}

    eval_ent_indices: Dict[str, List[Tuple[int, int]]] = {
        "correct_indices": [],
        "incorrect_indices": [],
        "partial_indices": [],
        "missed_indices": [],
        "spurious_indices": [],
    }

    # Create dicts to hold indices for correct/spurious/missing/etc examples
    evaluation_ent_indices = {
        "strict": deepcopy(eval_ent_indices),
        "ent_type": deepcopy(eval_ent_indices),
        "partial": deepcopy(eval_ent_indices),
        "exact": deepcopy(eval_ent_indices),
    }
    evaluation_agg_ent_indices = {e: deepcopy(evaluation_ent_indices) for e in tags}

    # <INJECTION> #
    confusion_matrix = np.zeros((len(tags), len(tags)), dtype=np.int_)
    label2id = {tag: idx for idx, tag in enumerate(tags)}

    # </INJECTION> #

    # keep track of entities that overlapped
    true_which_overlapped_with_pred = []

    # Subset into only the tags that we are interested in.
    # NOTE: we remove the tags we don't want from both the predicted and the
    # true entities. This covers the two cases where mismatches can occur:
    #
    # 1) Where the model predicts a tag that is not present in the true data
    # 2) Where there is a tag in the true data that the model is not capable of
    # predicting.

    # Strip the spans down to just start, end, label. Note that failing
    # to do this results in a bug. The exact cause is not clear.
    true_named_entities = [
        clean_entities(ent) for ent in true_named_entities if ent["label"] in tags
    ]
    pred_named_entities = [
        clean_entities(ent) for ent in pred_named_entities if ent["label"] in tags
    ]

    # Sort the lists to improve the speed of the overlap comparison
    true_named_entities.sort(key=lambda x: x["start"])
    pred_named_entities.sort(key=lambda x: x["end"])

    # go through each predicted named-entity
    for within_instance_index, pred in enumerate(pred_named_entities):
        found_overlap = False

        # Check each of the potential scenarios in turn. See
        # http://www.davidsbatista.net/blog/2018/05/09/Named_Entity_Evaluation/
        # for scenario explanation.

        # Scenario I: Exact match between true and pred
        if pred in true_named_entities:
            true_which_overlapped_with_pred.append(pred)
            evaluation["strict"]["correct"] += 1
            evaluation["ent_type"]["correct"] += 1
            evaluation["exact"]["correct"] += 1
            evaluation["partial"]["correct"] += 1
            evaluation_ent_indices["strict"]["correct_indices"].append(
                (instance_index, within_instance_index)
            )
            evaluation_ent_indices["ent_type"]["correct_indices"].append(
                (instance_index, within_instance_index)
            )
            evaluation_ent_indices["exact"]["correct_indices"].append(
                (instance_index, within_instance_index)
            )
            evaluation_ent_indices["partial"]["correct_indices"].append(
                (instance_index, within_instance_index)
            )

            # for the agg. by label results
            evaluation_agg_entities_type[pred["label"]]["strict"]["correct"] += 1
            evaluation_agg_entities_type[pred["label"]]["ent_type"]["correct"] += 1
            evaluation_agg_entities_type[pred["label"]]["exact"]["correct"] += 1
            evaluation_agg_entities_type[pred["label"]]["partial"]["correct"] += 1
            evaluation_agg_ent_indices[pred["label"]]["strict"][
                "correct_indices"
            ].append((instance_index, within_instance_index))
            evaluation_agg_ent_indices[pred["label"]]["ent_type"][
                "correct_indices"
            ].append((instance_index, within_instance_index))
            evaluation_agg_ent_indices[pred["label"]]["exact"][
                "correct_indices"
            ].append((instance_index, within_instance_index))
            evaluation_agg_ent_indices[pred["label"]]["partial"][
                "correct_indices"
            ].append((instance_index, within_instance_index))

            # <INJECTION> #
            confusion_matrix[label2id[pred["label"]], label2id[pred["label"]]] += 1
            # </INJECTION> #

        else:
            # check for overlaps with any of the true entities
            for true in true_named_entities:
                # Only enter this block if an overlap is possible
                if pred["end"] < true["start"]:
                    break

                # overlapping needs to take into account last token as well
                pred_range = range(pred["start"], pred["end"] + 1)
                true_range = range(true["start"], true["end"] + 1)

                # Scenario IV: Offsets match, but entity type is wrong
                if (
                    true["start"] == pred["start"]
                    and pred["end"] == true["end"]
                    and true["label"] != pred["label"]
                ):
                    # overall results
                    evaluation["strict"]["incorrect"] += 1
                    evaluation_ent_indices["strict"]["incorrect_indices"].append(
                        (instance_index, within_instance_index)
                    )
                    evaluation["ent_type"]["incorrect"] += 1
                    evaluation_ent_indices["ent_type"]["incorrect_indices"].append(
                        (instance_index, within_instance_index)
                    )
                    evaluation["partial"]["correct"] += 1
                    evaluation["exact"]["correct"] += 1

                    # aggregated by entity type results
                    evaluation_agg_entities_type[true["label"]]["strict"][
                        "incorrect"
                    ] += 1
                    evaluation_agg_ent_indices[true["label"]]["strict"][
                        "incorrect_indices"
                    ].append((instance_index, within_instance_index))
                    evaluation_agg_entities_type[true["label"]]["ent_type"][
                        "incorrect"
                    ] += 1
                    evaluation_agg_ent_indices[true["label"]]["ent_type"][
                        "incorrect_indices"
                    ].append((instance_index, within_instance_index))
                    evaluation_agg_entities_type[true["label"]]["partial"][
                        "correct"
                    ] += 1
                    evaluation_agg_entities_type[true["label"]]["exact"]["correct"] += 1

                    true_which_overlapped_with_pred.append(true)
                    found_overlap = True

                    # <INJECTION> #
                    confusion_matrix[
                        label2id[true["label"]], label2id[pred["label"]]
                    ] += 1
                    # </INJECTION> #
                    break

                # check for an overlap i.e. not exact boundary match, with true entities
                # overlaps with true entities must only count once
                if (
                    find_overlap(true_range, pred_range)
                    and true not in true_which_overlapped_with_pred
                ):
                    true_which_overlapped_with_pred.append(true)

                    # Scenario V: There is an overlap (but offsets do not match
                    # exactly), and the entity type is the same.
                    # 2.1 overlaps with the same entity type
                    if pred["label"] == true["label"]:
                        # overall results
                        evaluation["strict"]["incorrect"] += 1
                        evaluation_ent_indices["strict"]["incorrect_indices"].append(
                            (instance_index, within_instance_index)
                        )
                        evaluation["ent_type"]["correct"] += 1
                        evaluation["partial"]["partial"] += 1
                        evaluation_ent_indices["partial"]["partial_indices"].append(
                            (instance_index, within_instance_index)
                        )
                        evaluation["exact"]["incorrect"] += 1
                        evaluation_ent_indices["exact"]["incorrect_indices"].append(
                            (instance_index, within_instance_index)
                        )

                        # aggregated by entity type results
                        evaluation_agg_entities_type[true["label"]]["strict"][
                            "incorrect"
                        ] += 1
                        evaluation_agg_ent_indices[true["label"]]["strict"][
                            "incorrect_indices"
                        ].append((instance_index, within_instance_index))
                        evaluation_agg_entities_type[true["label"]]["ent_type"][
                            "correct"
                        ] += 1
                        evaluation_agg_entities_type[true["label"]]["partial"][
                            "partial"
                        ] += 1
                        evaluation_agg_ent_indices[true["label"]]["partial"][
                            "partial_indices"
                        ].append((instance_index, within_instance_index))
                        evaluation_agg_entities_type[true["label"]]["exact"][
                            "incorrect"
                        ] += 1
                        evaluation_agg_ent_indices[true["label"]]["exact"][
                            "incorrect_indices"
                        ].append((instance_index, within_instance_index))

                        found_overlap = True

                    else:
                        # Scenario VI: Entities overlap, but the entity type is
                        # different.

                        # overall results
                        evaluation["strict"]["incorrect"] += 1
                        evaluation_ent_indices["strict"]["incorrect_indices"].append(
                            (instance_index, within_instance_index)
                        )
                        evaluation["ent_type"]["incorrect"] += 1
                        evaluation_ent_indices["ent_type"]["incorrect_indices"].append(
                            (instance_index, within_instance_index)
                        )
                        evaluation["partial"]["partial"] += 1
                        evaluation_ent_indices["partial"]["partial_indices"].append(
                            (instance_index, within_instance_index)
                        )
                        evaluation["exact"]["incorrect"] += 1
                        evaluation_ent_indices["exact"]["incorrect_indices"].append(
                            (instance_index, within_instance_index)
                        )

                        # aggregated by entity type results
                        # Results against the true entity

                        evaluation_agg_entities_type[true["label"]]["strict"][
                            "incorrect"
                        ] += 1
                        evaluation_agg_ent_indices[true["label"]]["strict"][
                            "incorrect_indices"
                        ].append((instance_index, within_instance_index))
                        evaluation_agg_entities_type[true["label"]]["partial"][
                            "partial"
                        ] += 1
                        evaluation_agg_ent_indices[true["label"]]["partial"][
                            "partial_indices"
                        ].append((instance_index, within_instance_index))
                        evaluation_agg_entities_type[true["label"]]["ent_type"][
                            "incorrect"
                        ] += 1
                        evaluation_agg_ent_indices[true["label"]]["ent_type"][
                            "incorrect_indices"
                        ].append((instance_index, within_instance_index))
                        evaluation_agg_entities_type[true["label"]]["exact"][
                            "incorrect"
                        ] += 1
                        evaluation_agg_ent_indices[true["label"]]["exact"][
                            "incorrect_indices"
                        ].append((instance_index, within_instance_index))

                        # Results against the predicted entity
                        # evaluation_agg_entities_type[pred['label']]['strict']['spurious'] += 1
                        found_overlap = True

            # Scenario II: Entities are spurious (i.e., over-generated).
            if not found_overlap:
                # Overall results
                evaluation["strict"]["spurious"] += 1
                evaluation_ent_indices["strict"]["spurious_indices"].append(
                    (instance_index, within_instance_index)
                )
                evaluation["ent_type"]["spurious"] += 1
                evaluation_ent_indices["ent_type"]["spurious_indices"].append(
                    (instance_index, within_instance_index)
                )
                evaluation["partial"]["spurious"] += 1
                evaluation_ent_indices["partial"]["spurious_indices"].append(
                    (instance_index, within_instance_index)
                )
                evaluation["exact"]["spurious"] += 1
                evaluation_ent_indices["exact"]["spurious_indices"].append(
                    (instance_index, within_instance_index)
                )

                # Aggregated by entity type results

                # a over-generated entity with a valid tag should be
                # attributed to the respective tag only
                if pred["label"] in tags:
                    spurious_tags = [pred["label"]]
                else:
                    # NOTE: when pred.e_type is not found in valid tags
                    # or when it simply does not appear in the test set, then it is
                    # spurious, but it is not clear where to assign it at the tag
                    # level. In this case, it is applied to all target_tags
                    # found in this example. This will mean that the sum of the
                    # evaluation_agg_entities will not equal evaluation.

                    spurious_tags = tags

                for true in spurious_tags:
                    evaluation_agg_entities_type[true]["strict"]["spurious"] += 1
                    evaluation_agg_ent_indices[true]["strict"][
                        "spurious_indices"
                    ].append((instance_index, within_instance_index))
                    evaluation_agg_entities_type[true]["ent_type"]["spurious"] += 1
                    evaluation_agg_ent_indices[true]["ent_type"][
                        "spurious_indices"
                    ].append((instance_index, within_instance_index))
                    evaluation_agg_entities_type[true]["partial"]["spurious"] += 1
                    evaluation_agg_ent_indices[true]["partial"][
                        "spurious_indices"
                    ].append((instance_index, within_instance_index))
                    evaluation_agg_entities_type[true]["exact"]["spurious"] += 1
                    evaluation_agg_ent_indices[true]["exact"][
                        "spurious_indices"
                    ].append((instance_index, within_instance_index))

    # Scenario III: Entity was missed entirely.
    for within_instance_index, true in enumerate(true_named_entities):
        if true in true_which_overlapped_with_pred:
            continue

        # overall results
        evaluation["strict"]["missed"] += 1
        evaluation_ent_indices["strict"]["missed_indices"].append(
            (instance_index, within_instance_index)
        )
        evaluation["ent_type"]["missed"] += 1
        evaluation_ent_indices["ent_type"]["missed_indices"].append(
            (instance_index, within_instance_index)
        )
        evaluation["partial"]["missed"] += 1
        evaluation_ent_indices["partial"]["missed_indices"].append(
            (instance_index, within_instance_index)
        )
        evaluation["exact"]["missed"] += 1
        evaluation_ent_indices["exact"]["missed_indices"].append(
            (instance_index, within_instance_index)
        )

        # for the agg. by label
        evaluation_agg_entities_type[true["label"]]["strict"]["missed"] += 1
        evaluation_agg_ent_indices[true["label"]]["strict"]["missed_indices"].append(
            (instance_index, within_instance_index)
        )
        evaluation_agg_entities_type[true["label"]]["ent_type"]["missed"] += 1
        evaluation_agg_ent_indices[true["label"]]["ent_type"]["missed_indices"].append(
            (instance_index, within_instance_index)
        )
        evaluation_agg_entities_type[true["label"]]["partial"]["missed"] += 1
        evaluation_agg_ent_indices[true["label"]]["partial"]["missed_indices"].append(
            (instance_index, within_instance_index)
        )
        evaluation_agg_entities_type[true["label"]]["exact"]["missed"] += 1
        evaluation_agg_ent_indices[true["label"]]["exact"]["missed_indices"].append(
            (instance_index, within_instance_index)
        )

    # Compute 'possible', 'actual' according to SemEval-2013 Task 9.1 on the
    # overall results, and use these to calculate precision and recall.
    for eval_type in evaluation:
        evaluation[eval_type] = compute_actual_possible(evaluation[eval_type])

    # Compute 'possible', 'actual', and precision and recall on entity level
    # results. Start by cycling through the accumulated results.
    for entity_type, entity_level in evaluation_agg_entities_type.items():
        # Cycle through the evaluation types for each dict containing entity
        # level results.

        for eval_type in entity_level:
            evaluation_agg_entities_type[entity_type][eval_type] = (
                compute_actual_possible(entity_level[eval_type])
            )

    return (
        evaluation,
        evaluation_agg_entities_type,
        evaluation_ent_indices,
        evaluation_agg_ent_indices,
        confusion_matrix,
    )


def compute_actual_possible(results: Dict) -> Dict:
    """
    Takes a result dict that has been output by compute metrics.
    Returns the results' dict with actual, possible populated.

    When the results dicts is from partial or ent_type metrics, then
    partial_or_type=True to ensure the right calculation is used for
    calculating precision and recall.
    """

    correct = results["correct"]
    incorrect = results["incorrect"]
    partial = results["partial"]
    missed = results["missed"]
    spurious = results["spurious"]

    # Possible: number annotations in the gold-standard which contribute to the
    # final score
    possible = correct + incorrect + partial + missed

    # Actual: number of annotations produced by the NER system
    actual = correct + incorrect + partial + spurious

    results["actual"] = actual
    results["possible"] = possible

    return results


def compute_precision_recall(results: Dict, partial_or_type: bool = False) -> Dict:
    """
    Takes a result dict that has been output by compute metrics.
    Returns the results' dict with precision and recall populated.

    When the results dicts is from partial or ent_type metrics, then
    partial_or_type=True to ensure the right calculation is used for
    calculating precision and recall.
    """

    actual = results["actual"]
    possible = results["possible"]
    partial = results["partial"]
    correct = results["correct"]

    if partial_or_type:
        precision = (correct + 0.5 * partial) / actual if actual > 0 else 0
        recall = (correct + 0.5 * partial) / possible if possible > 0 else 0

    else:
        precision = correct / actual if actual > 0 else 0
        recall = correct / possible if possible > 0 else 0

    results["precision"] = precision
    results["recall"] = recall
    results["f1"] = (
        2 * (precision * recall) / (precision + recall)
        if (precision + recall) > 0
        else 0
    )

    return results


def compute_precision_recall_wrapper(results: Dict) -> Dict:
    """
    Wraps the compute_precision_recall function and runs on a dict of results
    """

    results_a = {
        key: compute_precision_recall(value, True)
        for key, value in results.items()
        if key in ["partial", "ent_type"]
    }
    results_b = {
        key: compute_precision_recall(value)
        for key, value in results.items()
        if key in ["strict", "exact"]
    }

    results = {**results_a, **results_b}

    return results


def clean_entities(ent: Dict) -> Dict:
    """
    Returns just the useful keys if additional keys are present in the entity
    dict.

    This may happen if passing a list of spans directly from prodigy, which
    typically may include 'token_start' and 'token_end'.
    """
    return {"start": ent["start"], "end": ent["end"], "label": ent["label"]}


def summary_report_ent(  # pylint: disable=too-many-locals
    results_agg_entities_type: Dict, scenario: str = "strict", digits: int = 2
) -> str:
    if scenario not in {"strict", "ent_type", "partial", "exact"}:
        raise Exception(
            "Invalid scenario: must be one of 'strict', 'ent_type', 'partial', 'exact'"
        )

    target_names = sorted(results_agg_entities_type.keys())
    headers = [
        "correct",
        "incorrect",
        "partial",
        "missed",
        "spurious",
        "precision",
        "recall",
        "f1-score",
    ]
    rows = [headers]

    for ent_type, results in sorted(results_agg_entities_type.items()):
        for k, v in results.items():
            if k != scenario:
                continue
            rows.append(
                [
                    ent_type,
                    v["correct"],
                    v["incorrect"],
                    v["partial"],
                    v["missed"],
                    v["spurious"],
                    v["precision"],
                    v["recall"],
                    v["f1"],
                ]
            )

    name_width = max(len(cn) for cn in target_names)
    width = max(name_width, digits)
    head_fmt = "{:>{width}s} " + " {:>11}" * len(headers)
    report = head_fmt.format("", *headers, width=width)
    report += "\n\n"
    row_fmt = "{:>{width}s} " + " {:>11}" * 5 + " {:>11.{digits}f}" * 3 + "\n"

    for row in rows[1:]:
        report += row_fmt.format(*row, width=width, digits=digits)

    return report


def summary_report_overall(results: Dict, digits: int = 2) -> str:
    headers = [
        "correct",
        "incorrect",
        "partial",
        "missed",
        "spurious",
        "precision",
        "recall",
        "f1-score",
    ]
    rows = [headers]

    for k, v in results.items():
        rows.append(
            [
                k,
                v["correct"],
                v["incorrect"],
                v["partial"],
                v["missed"],
                v["spurious"],
                v["precision"],
                v["recall"],
                v["f1"],
            ]
        )

    target_names = sorted(results.keys())
    name_width = max(len(cn) for cn in target_names)
    width = max(name_width, digits)
    head_fmt = "{:>{width}s} " + " {:>11}" * len(headers)
    report = head_fmt.format("", *headers, width=width)
    report += "\n\n"
    row_fmt = "{:>{width}s} " + " {:>11}" * 5 + " {:>11.{digits}f}" * 3 + "\n"

    for row in rows[1:]:
        report += row_fmt.format(*row, width=width, digits=digits)

    return report


def summary_report_ents_indices(
    evaluation_agg_indices: Dict, error_schema: str, preds: Optional[List] = [[]]
) -> str:
    """
    Usage: print(summary_report_ents_indices(evaluation_agg_indices, 'partial', preds))
    """
    report = ""
    for entity_type, entity_results in evaluation_agg_indices.items():
        report += f"\nEntity Type: {entity_type}\n"
        error_data = entity_results[error_schema]
        report += f"  Error Schema: '{error_schema}'\n"
        for category, indices in error_data.items():
            category_name = category.replace("_", " ").capitalize()
            report += f"    ({entity_type}) {category_name}:\n"
            if indices:
                for instance_index, entity_index in indices:
                    if preds != [[]]:
                        pred = preds[instance_index][entity_index]  # type: ignore
                        prediction_info = f"Label={pred['label']}, Start={pred['start']}, End={pred['end']}"
                        report += f"      - Instance {instance_index}, Entity {entity_index}: {prediction_info}\n"
                    else:
                        report += f"      - Instance {instance_index}, Entity {entity_index}\n"
            else:
                report += "      - None\n"
    return report


def summary_report_overall_indices(
    evaluation_indices: Dict, error_schema: str, preds: Optional[List] = [[]]
) -> str:
    """
    Usage: print(summary_report_overall_indices(evaluation_indices, 'partial', preds))
    """
    report = ""
    assert (
        error_schema in evaluation_indices
    ), f"Error schema '{error_schema}' not found in the results."

    error_data = evaluation_indices[error_schema]
    report += f"Indices for error schema '{error_schema}':\n\n"

    for category, indices in error_data.items():
        category_name = category.replace("_", " ").capitalize()
        report += f"{category_name}:\n"
        if indices:
            for instance_index, entity_index in indices:
                if preds != [[]]:
                    # Retrieve the corresponding prediction
                    pred = preds[instance_index][entity_index]  # type: ignore
                    prediction_info = f"Label={pred['label']}, Start={pred['start']}, End={pred['end']}"
                    report += f"  - Instance {instance_index}, Entity {entity_index}: {prediction_info}\n"
                else:
                    report += f"  - Instance {instance_index}, Entity {entity_index}\n"
        else:
            report += "  - None\n"
        report += "\n"

    return report



from typing import Any, Optional
import openai
from .configs.llm_config import LlmClientConfig


class ModelEndpointWrapper:
    """Wrapper for OpenAI-compatible LLM API endpoints.

    Provides a callable interface for making streaming chat completion requests
    to LLM APIs. Supports customizable generation parameters per-request while
    maintaining sensible defaults from configuration.

    Attributes:
        config_cls: The configuration class used for this wrapper.
        client: The OpenAI client instance.
        system_prompt: Default system prompt for conversations.
        model_name: Model identifier for API requests.
        default_temperature: Default sampling temperature.
        default_presence: Default presence penalty.
        default_frequency_penalty: Default frequency penalty.
        num_concurrent_requests: Maximum concurrent API requests allowed.

    Example:
        >>> config = LlmClientConfig(api_key="your-key", model_id="gpt-4")
        >>> llm = ModelEndpointWrapper(config)
        >>> response = llm("What is machine learning?")
        >>> print(response)
    """

    config_cls = LlmClientConfig

    def __init__(
        self,
        config: Optional[LlmClientConfig] = None,
    ) -> None:
        """Initialize the model endpoint wrapper.

        Args:
            config: Configuration for the LLM client. If None, uses default
                configuration with values from environment variables.
        """
        if config is None:
            config = LlmClientConfig()

        self.client = openai.OpenAI(base_url=config.base_url, api_key=config.api_key)
        self.system_prompt = config.default_system_prompt
        self.model_name = config.model_id
        self.default_temperature = config.default_temperature
        self.default_presence = config.default_presence_penalty
        self.default_frequency_penalty = config.default_frequency_penalty
        self.num_concurrent_requests = config.num_concurrent_requests

    def __call__(
        self,
        prompt,
        system_prompt=None,
        temperature=None,
        presence_penalty=None,
        frequency_penalty=None,
    ) -> Any:
        """Send a prompt to the LLM and return the generated response.

        Makes a streaming chat completion request and concatenates the response
        chunks into a single string.

        Args:
            prompt: The user message to send to the model.
            system_prompt: Override the default system prompt for this request.
            temperature: Sampling temperature (0.0-1.0). Higher values produce
                more random outputs.
            presence_penalty: Penalty for repeating topics already mentioned.
            frequency_penalty: Penalty for repeating the same tokens.

        Returns:
            The complete generated response as a string.
        """
        params = {
            "temperature": (
                self.default_temperature if temperature is None else temperature
            ),
            "presence_penalty": (
                self.default_presence if presence_penalty is None else presence_penalty
            ),
            "frequency_penalty": (
                self.default_frequency_penalty
                if frequency_penalty is None
                else frequency_penalty
            ),
        }

        stream = self.client.chat.completions.create(
            messages=[
                {
                    "role": "system",
                    "content": system_prompt or self.system_prompt,
                },
                {
                    "role": "user",
                    "content": prompt,
                },
            ],
            stream=True,
            model=self.model_name,
            **params,
        )
        result = []
        for chunk in stream:
            result.append(chunk.choices[0].delta.content or "")
        return "".join(result)



from dataclasses import dataclass, field
from .base_config import BaseMethodConfig
from ..data_types import OodMethod
from ..finetuning_training_args import FinetuningTrainingArguments


@dataclass
class FinetunerConfig(BaseMethodConfig):
    """Configuration for transformer finetuning method.

    Attributes:
        training_arguments: HuggingFace training arguments with
            hyperparameter search space configuration.
        num_hpo_trials: Number of Optuna trials for hyperparameter
            optimization. Set to 0 to disable HPO.
    """

    training_arguments: FinetuningTrainingArguments = field(
        default_factory=FinetuningTrainingArguments  # type: ignore
    )
    num_hpo_trials: int = 10
    ood_method: OodMethod = OodMethod.NONE



from typing import List, Optional
import outlines
import pandas as pd

from .model_endpoint import ModelEndpointWrapper
from .processor import Processor
from .configs.data_generator_config import DataGeneratorConfig


class DataGenerator(Processor):
    config_cls = DataGeneratorConfig

    def __init__(
        self,
        prompt: outlines.prompts.Prompt,
        client: ModelEndpointWrapper,
        config: Optional[DataGeneratorConfig] = None,
    ):
        if config is None:
            config = DataGeneratorConfig()
        super().__init__(prompt=prompt, client=client, **config.__dict__)

    def _aggregate_batches(self, batches: List[pd.DataFrame]) -> pd.DataFrame:
        """
        Aggregate a list of DataFrame batches into a single DataFrame, renaming columns and removing duplicates.

        Parameters:
        batches (List[pd.DataFrame]): A list of DataFrame batches to be aggregated.

        Returns:
        pd.DataFrame: A single aggregated DataFrame with duplicates removed.
        """
        all_dfs = []
        for df in batches:
            df = df.rename({"Text": "text"}, axis=1)
            all_dfs.append(df)
        aggregated = pd.concat(all_dfs, ignore_index=True).drop_duplicates("text")
        return aggregated



import logging
from typing import List, Optional, Type

from datasets import Dataset

from .. import methods
from .abstract_pipeline import (
    AbstractInferencePipeline,
    AbstractTrainingPipeline,
    get_method_from_string,
)
from ..constants import OOD_METHOD_MAP, TRAIN_KEY
from ..data.ner_data_provider import SimpleNerDataProvider
from ..data_quality import LabelAggregation, Retag, Uncertainty
from ..data_quality.data_types import DatasetEvaluatorOutput
from ..data_quality.dynamic_finetuner import TokenClassDynamicTuner
from ..methods.data_types import TokenClassifierOutput
from ..method_resolver import (
    count_labels_ner,
    resolve_ner_method,
)

log = logging.getLogger(__name__)


class TokenClassificationTrainingPipeline(AbstractTrainingPipeline):
    """Training pipeline for token classification (NER) tasks.

    Supports standard token classification finetuning approach.

    Example:
        >>> from open_autonlu.auto_classes import TokenClassificationTrainingPipeline
        >>> from open_autonlu.methods.data_types import SaveFormat
        >>>
        >>> pipeline = TokenClassificationTrainingPipeline(
        ...     train_path="train.json",
        ...     test_path="test.json"
        ... )
        >>> result = pipeline.train()
        >>> pipeline.save("./model", SaveFormat.ONNX)
    """

    data_provider_cls = SimpleNerDataProvider
    data_quality_tuner_cls = TokenClassDynamicTuner
    data_quality_evaluators_classes = [Retag, Uncertainty, LabelAggregation]
    training_method_cls: Type[methods.Method]
    task_type = "token_classification"

    def diagnose(self) -> Optional[DatasetEvaluatorOutput]:
        """Run data quality diagnostics on the NER training dataset.

        Evaluates the training data using Retag, Uncertainty, and
        LabelAggregation evaluators to identify problematic samples.

        Returns:
            DatasetEvaluatorOutput with evaluation results.
        """
        evaluation_result: DatasetEvaluatorOutput = self.evaluate_training_data()
        return evaluation_result

    def _synergize_data_and_method(self) -> tuple[Type[methods.Method], Dataset]:
        """Analyze NER data and select the appropriate training method.

        Counts entity label distribution and selects the training method
        based on the maximum label count in the dataset.

        Returns:
            Tuple of (selected method class, training dataset).
        """
        training_data = self.splits[TRAIN_KEY]
        label_counts, _ = count_labels_ner(training_data)
        method_name = resolve_ner_method(label_counts.max())
        return get_method_from_string(OOD_METHOD_MAP[method_name]), training_data


class TokenClassificationInferenceManager(AbstractInferencePipeline):
    """Inference pipeline for token classification (NER) models.

    Loads a trained NER model and provides entity extraction
    capabilities for new text inputs.

    Example:
        >>> from open_autonlu.auto_classes import TokenClassificationInferenceManager
        >>>
        >>> inferer = TokenClassificationInferenceManager("./model")
        >>> results = inferer.predict(["John works at Google"], batch_size=1)
        >>> for r in results:
        ...     print(r.entities)
    """

    def predict(
        self, texts: List[str], batch_size: int, return_all_hypotheses: bool = False
    ) -> List[TokenClassifierOutput]:
        """Run NER inference on a list of texts.

        Args:
            texts: List of text strings to extract entities from.
            batch_size: Number of texts to process in each batch.
            return_all_hypotheses: Not supported for token classification.
                Will be ignored with a debug warning.

        Returns:
            List of TokenClassifierOutput, one per input text, containing
            extracted entities with their spans and labels.
        """
        if return_all_hypotheses:
            log.debug(
                "return_all_hypotheses is not supported for Token Classification and will be ignored."
            )
        return self.inferer.predict(
            texts,
            batch_size=batch_size,
        )



"""
Example: NER pipeline.
"""

import torch

from open_autonlu.auto_classes import (
    TokenClassificationTrainingPipeline,
)

TRAIN_PATH = "examples/test_data/noise_n_shot_data_ner/train.json"
TEST_PATH = "examples/test_data/noise_n_shot_data_ner/test.json"


def main():
    if not torch.cuda.is_available() and not torch.backends.mps.is_available():
        print("Warning: no GPU; training may be slow.")

    pipeline = TokenClassificationTrainingPipeline(
        TRAIN_PATH,
        TEST_PATH,
        config_overrides={"language": "en"},
    )
    _ = pipeline.diagnose()
    training_artifact_info = pipeline.train()
    print("Selected method:", pipeline.training_method_cls.__name__)
    if training_artifact_info.test_metrics:
        print("Test F1:", training_artifact_info.test_metrics.f1)

    # Optional: save and run inference
    # pipeline.save("saved_ner_torch", SaveFormat.TORCH)
    # pipeline.save("saved_ner_onnx", SaveFormat.ONNX)
    # manager = TokenClassificationInferenceManager("saved_ner_torch")
    # sample_texts = ["Sample query one.", "Another example text."]
    # results = manager.predict(sample_texts, batch_size=32)
    # for r in results:
    #     print(r.text, [(e.label, e.text) for e in r.labels])


if __name__ == "__main__":
    main()



from typing import NamedTuple

import spacy
from datasets import Dataset

SUPPORTED_LANGUAGES = ("ru", "en")

_spacy_models: dict[str, spacy.language.Language] = {}


def _get_spacy_model(language: str) -> spacy.language.Language:
    if language not in _spacy_models:
        _spacy_models[language] = spacy.blank(language)
    return _spacy_models[language]


class Token(NamedTuple):
    start: int
    stop: int
    text: str


def tokenize_with_offsets(text: str, language: str = "en") -> list[Token]:
    """Tokenize text and return list of (start, stop, text) for each token."""
    if language not in SUPPORTED_LANGUAGES:
        raise ValueError(
            f"Unsupported language: '{language}'. Supported: {SUPPORTED_LANGUAGES}"
        )

    nlp = _get_spacy_model(language)
    doc = nlp(text)
    return [
        Token(
            start=token.idx if token.idx is not None else 0,
            stop=token.idx + len(token.text)
            if token.idx is not None
            else len(token.text),
            text=token.text,
        )
        for token in doc
    ]


def align_labels_with_tokens(labels, word_ids, b_to_i_label, label_all_tokens=True):
    new_labels = []
    current_word = None
    for word_id in word_ids:
        if word_id != current_word:
            # Start of a new word!
            current_word = word_id
            label = -100 if word_id is None else labels[word_id]
            new_labels.append(label)
        elif word_id is None:
            # Special token
            new_labels.append(-100)
        else:
            # Same word as previous token
            label = labels[word_id]
            # If the label is B-XXX we change it to I-XXX
            if label_all_tokens:
                new_labels.append(b_to_i_label[label])
            else:
                new_labels.append(-100)

    return new_labels


def convert_offsets_to_bio(
    text: str,
    labels: list,
    label_key: str = "label",
    language: str = "en",
) -> tuple[list[str], list[str]]:
    substrings = tokenize_with_offsets(text, language=language)
    tokens: list[str] = []
    bio_tags: list[str] = []

    def find_entity_for_token(t_start: int, t_end: int):
        for ent in labels:
            ent_start = ent["start"]
            ent_end = ent["end"]
            if not (t_end <= ent_start or t_start >= ent_end):
                return ent
        return None

    last_label = None

    for token in substrings:
        start, end = token.start, token.stop
        tokens.append(token.text)

        ent = find_entity_for_token(start, end)
        if ent is None:
            bio_tags.append("O")
            last_label = None
            continue

        label = ent[label_key]
        prefix = "I-" if last_label == label else "B-"
        bio_tags.append(prefix + label)
        last_label = label

    return tokens, bio_tags


def convert_bio_to_spans(
    text: str,
    tokens: list[str],
    bio_tags: list[str],
) -> list[dict]:
    """Convert token-level BIO tags back to character-offset spans"""
    if not tokens or not bio_tags or len(tokens) != len(bio_tags):
        return []

    pos = 0
    token_positions = []
    for t in tokens:
        idx = text.find(t, pos)
        if idx == -1:
            idx = text.find(t)
        if idx == -1:
            return []
        token_positions.append(Token(start=idx, stop=idx + len(t), text=t))
        pos = idx + len(t)

    spans = []
    i = 0
    while i < len(bio_tags):
        tag = bio_tags[i] if isinstance(bio_tags[i], str) else str(bio_tags[i])
        if tag.startswith("B-"):
            label = tag[2:]
            start = token_positions[i].start
            j = i + 1
            while j < len(bio_tags):
                next_tag = (
                    bio_tags[j] if isinstance(bio_tags[j], str) else str(bio_tags[j])
                )
                if next_tag != f"I-{label}":
                    break
                j += 1
            end = token_positions[j - 1].stop
            spans.append({"start": start, "end": end, "label": label})
            i = j
        else:
            i += 1
    return spans


def anc_label_is_present(dataset: Dataset) -> bool:
    return "anc_label" in dataset.column_names


def pop_default(input_list, index=-1, default_value=None):
    try:
        return input_list.pop(index)
    except IndexError:
        return default_value



import logging
from collections import Counter
from typing import Any, Dict, Optional

import pandas as pd
from datasets import Dataset

from open_autonlu.llm_pipelines import DataGenerationPipeline, analyze_domain_and_labels
from open_autonlu.llm_pipelines.prompts import get_prompts

from .constants import (
    ANCSETFIT_NAME,
    FINETUNING_NAME,
    LLM_AUGMENTATION_THRESHOLD,
    MAX_ANC_SET_FIT_SIZE,
    MAX_SET_FIT_SIZE,
    MIN_ANC_SET_FIT_SIZE,
    SETFIT_NAME,
    TOKEN_CLF_FINETUNING,
    UPSAMPLE_AUGMENTEX,
    UPSAMPLE_LLM,
)
from .data import WordCharAugmentation

log = logging.getLogger(__name__)


def downsample(
    dataset: Dataset, label_counts: pd.DataFrame, downsample_to: int
) -> Dataset:
    """
    Downsamples a given dataset to the specified size for each class.

    Args:
        dataset (Dataset): The input dataset.
        label_counts (pd.DataFrame): A DataFrame containing the count of each label in the dataset.
        downsample_to (int): The target size for each class.

    Returns:
        Dataset: The downsampled dataset.
    """
    df: pd.DataFrame = dataset.to_pandas()  # type: ignore
    labels_to_downsample = (
        label_counts.where(lambda x: x > downsample_to, pd.NA).dropna().index.tolist()
    )
    sub_dfs = []
    for label_name, sub_df in df.groupby("label"):
        if label_name in labels_to_downsample:
            sub_dfs.append(sub_df.sample(downsample_to, random_state=42))
        else:
            sub_dfs.append(sub_df)
    df = pd.concat(sub_dfs, ignore_index=True)
    return Dataset.from_pandas(df, preserve_index=False)


def upsample(
    dataset: Dataset,
    label_counts: pd.DataFrame,
    upsample_to: int,
    upsample_method: str = UPSAMPLE_AUGMENTEX,
    llm_config_overrides: Optional[Dict[str, Any]] = None,
    domain_desc: Optional[str] = None,
    label_descriptions: Optional[Dict[str, str]] = None,
    language: str = "en",
) -> Dataset:
    """
    Upsamples a given dataset to the specified size for each class using the specified method.

    Args:
        dataset (Dataset): The input dataset.
        label_counts (pd.DataFrame): A DataFrame containing the count of each label in the dataset.
        upsample_to (int): The target size for each class.
        upsample_method (str, optional): The method to use for upsampling. Defaults to UPSAMPLE_AUGMENTEX.
        llm_config_overrides: Optional config overrides for LLM augmentation.
        domain_desc: Optional domain description for LLM augmentation.
        label_descriptions: Optional label descriptions for LLM augmentation.

    Returns:
        Dataset: The upsampled dataset.
    """
    df: pd.DataFrame = dataset.to_pandas()  # type: ignore
    low_resource_classes = [
        str(label)  # Ensure labels are strings
        for label in label_counts.where(lambda x: x < upsample_to, pd.NA)
        .dropna()
        .index.tolist()
    ]
    # Ensure dataset labels are strings for consistent comparison
    if df["label"].dtype != "object":
        df = df.copy()
        df["label"] = df["label"].astype(str)
    if upsample_method == UPSAMPLE_AUGMENTEX:
        df = _augment_with_augmentex(df, low_resource_classes, upsample_to)
    elif upsample_method == UPSAMPLE_LLM:
        df = _augment_with_llm(
            dataset_df=df,
            label_counts=label_counts,
            upsample_to=upsample_to,
            llm_config_overrides=llm_config_overrides or {},
            domain_desc=domain_desc,
            label_descriptions=label_descriptions,
            language=language,
        )

    return Dataset.from_pandas(df, preserve_index=False)


def _augment_with_augmentex(
    df: pd.DataFrame, low_resource_classes: list[str], upsample_to: int
) -> pd.DataFrame:
    augmentation = WordCharAugmentation()
    sub_dfs = []
    for label_name, sub_df in df.groupby("label"):
        # Ensure label_name is string for comparison
        label_name_str = str(label_name)
        if label_name_str in low_resource_classes:
            augmented_pool = []
            for _, row in sub_df.iterrows():
                augmented_texts = augmentation(row["text"])
                for text in augmented_texts:
                    clone = row.copy()
                    clone["text"] = text
                    augmented_pool.append(clone)
            num_samples_to_aug = upsample_to - sub_df.shape[0]
            augmented_df = pd.DataFrame(augmented_pool).drop_duplicates(subset=["text"])
            if num_samples_to_aug > augmented_df.shape[0]:
                if augmented_df.shape[0] > 0:
                    sub_dfs.append(pd.concat([sub_df, augmented_df]))
                else:
                    sub_dfs.append(sub_df)
            else:
                augmented_df = augmented_df.sample(
                    n=num_samples_to_aug, random_state=42
                )
                sub_dfs.append(pd.concat([sub_df, augmented_df]))
        else:
            sub_dfs.append(sub_df)
    return pd.concat(sub_dfs, ignore_index=True)


def _augment_with_llm(
    dataset_df: pd.DataFrame,
    label_counts: pd.DataFrame,
    upsample_to: int,
    llm_config_overrides: Dict[str, Any],
    domain_desc: Optional[str],
    label_descriptions: Optional[Dict[str, str]],
    language: str = "en",
) -> pd.DataFrame:
    threshold = llm_config_overrides.get("threshold", LLM_AUGMENTATION_THRESHOLD)
    classes_to_augment = [
        str(label)
        for label, stats in label_counts.iterrows()
        if stats["text"] < threshold and str(label).lower() != "outofscope"
    ]

    if not classes_to_augment:
        log.info("No classes requiring LLM augmentation")
        return dataset_df

    if dataset_df["label"].dtype != "object":
        dataset_df = dataset_df.copy()
        dataset_df["label"] = dataset_df["label"].astype(str)

    domain_desc, label_descriptions = _ensure_domain_context(
        dataset_df,
        domain_desc,
        label_descriptions,
        llm_config_overrides.get("use_domain_analysis", False),
        llm_config_overrides.get("config_overrides"),
        language=language,
    )

    prompt = llm_config_overrides.get("prompt") or get_prompts(language).generate_texts
    llm_pipeline_overrides = llm_config_overrides.get("config_overrides")

    llm_classes_data = {}
    fallback_rows = []

    for label in classes_to_augment:
        sub_df = dataset_df[dataset_df["label"] == label]
        current_size = sub_df.shape[0]
        num_needed = max(0, upsample_to - current_size)
        if num_needed <= 0:
            continue

        if current_size < 2:
            log.warning(
                "Label '%s' has less than 2 examples. Using character-level augmentation.",
                label,
            )
            fallback_df = _augment_with_augmentex(
                sub_df, low_resource_classes=[label], upsample_to=upsample_to
            )
            new_rows = fallback_df.iloc[sub_df.shape[0] :]
            if not new_rows.empty:
                fallback_rows.append(new_rows)
            continue

        llm_classes_data[label] = {
            "reference_data": sub_df,
            "current_size": current_size,
            "num_needed": num_needed,
        }

    augmented_rows = []

    if llm_classes_data:
        all_reference_data = pd.concat(
            [data["reference_data"] for data in llm_classes_data.values()],
            ignore_index=True,
        )

        pipeline = DataGenerationPipeline(
            reference_data=all_reference_data,
            text_column="text",
            label_column="label",
            prompt=prompt,
            config_overrides=llm_pipeline_overrides,
            domain_desc=domain_desc,
            label_descriptions=label_descriptions,
            language=language,
        )

        all_generated_texts_per_class = {label: [] for label in llm_classes_data.keys()}
        max_attempts = llm_config_overrides.get("max_attempts", 10)
        attempt = 0

        log.info("Generating data for %d classes in parallel", len(llm_classes_data))

        while attempt < max_attempts:
            classes_still_needing = {
                label: data
                for label, data in llm_classes_data.items()
                if len(all_generated_texts_per_class[label]) < data["num_needed"]
            }

            if not classes_still_needing:
                break

            max_num_needed = max(
                data["num_needed"] - len(all_generated_texts_per_class[label])
                for label, data in classes_still_needing.items()
            )
            min_request_size = llm_config_overrides.get("min_request_size", 20)
            request_size = max(int(max_num_needed * 1.2) + 1, int(min_request_size))
            num_shot = min(
                llm_config_overrides.get("num_shot", 5),
                min(data["current_size"] for data in classes_still_needing.values()),
            )

            generation_kwargs = {
                "num_generations_per_class": request_size,
                "num_shot": num_shot,
            }

            try:
                generated_dataset = pipeline.generate(generation_kwargs)

                if generated_dataset is None:
                    log.warning("Failed to generate data on attempt %d", attempt + 1)
                    attempt += 1
                    continue

                generated_df = generated_dataset.to_pandas()
                if generated_df.empty:
                    attempt += 1
                    continue

                for label, data in classes_still_needing.items():
                    label_generated = generated_df[generated_df["label"] == label]
                    num_needed = data["num_needed"]
                    current_generated = all_generated_texts_per_class[label]

                    if not label_generated.empty:
                        new_texts = label_generated["text"].tolist()
                        for text in new_texts:
                            if text not in current_generated:
                                current_generated.append(text)
                                if len(current_generated) >= num_needed:
                                    break

                attempt += 1

            except Exception as e:
                log.warning("Error during generation on attempt %d: %s", attempt + 1, e)
                attempt += 1
                continue

        for label, data in llm_classes_data.items():
            generated_texts = all_generated_texts_per_class[label]
            num_needed = data["num_needed"]

            if len(generated_texts) < num_needed:
                log.warning(
                    "Generated only %d/%d examples for label '%s' after %d attempts",
                    len(generated_texts),
                    num_needed,
                    label,
                    attempt,
                )

            if generated_texts:
                generated_df_class = pd.DataFrame(
                    {"text": generated_texts[:num_needed], "label": label}
                )
                augmented_rows.append(generated_df_class)

    # Combine LLM-generated and fallback-generated rows
    all_augmented_rows = augmented_rows + fallback_rows

    if all_augmented_rows:
        dataset_df = pd.concat([dataset_df] + all_augmented_rows, ignore_index=True)
    return dataset_df


def _ensure_domain_context(
    dataset_df: pd.DataFrame,
    domain_desc: Optional[str],
    label_descriptions: Optional[Dict[str, str]],
    use_domain_analysis: bool,
    config_overrides: Optional[Dict[str, Any]],
    language: str = "en",
) -> tuple[Optional[str], Optional[Dict[str, str]]]:
    if domain_desc is not None and label_descriptions is not None:
        return domain_desc, label_descriptions

    if not use_domain_analysis:
        return domain_desc, label_descriptions

    try:
        label_counts = dataset_df.groupby("label").size()
        min_label_count = label_counts.min() if len(label_counts) > 0 else 0
        examples_per_label = min(7, min_label_count) if min_label_count > 0 else 5

        domain_desc, label_descriptions = analyze_domain_and_labels(
            reference_data=dataset_df,
            text_column="text",
            label_column="label",
            examples_per_label=examples_per_label,
            config_overrides=config_overrides,
            language=language,
        )
    except Exception as exc:  # pylint: disable=broad-except
        log.warning("Failed to perform domain analysis: %s", exc, exc_info=True)
    return domain_desc, label_descriptions


def count_labels(dataset: Dataset) -> tuple[pd.DataFrame, int]:
    """
    Counts the occurrence of each label in a given dataset.

    Args:
        dataset (Dataset): The input dataset.

    Returns:
        tuple[pd.DataFrame, int]: A DataFrame containing the count of each label and the minimum class size.
    """
    df: pd.DataFrame = dataset.to_pandas()  # type: ignore
    label_counts = df.groupby("label").agg("count")
    min_class_size = label_counts.agg("min")["text"]
    return label_counts, min_class_size


def validate_min_size(label_counts: pd.DataFrame, min_class_size: int):
    """
    Validates the minimum class size.

    Args:
        label_counts (pd.DataFrame): A DataFrame containing the count of each label.
        min_class_size (int): The minimum class size.

    Raises:
        ValueError: If the minimum class size is less than MIN_ANC_SET_FIT_SIZE for any class.
    """
    if min_class_size < MIN_ANC_SET_FIT_SIZE:
        not_sufficient_list = (
            label_counts.where(lambda x: x < MIN_ANC_SET_FIT_SIZE, pd.NA)
            .dropna()
            .index.tolist()
        )
        raise ValueError(
            f"There are not enough data for labels [{','.join(map(str, not_sufficient_list))}]"
        )


def resolve_method(min_class_size: int) -> str:
    """
    Resolves the method to use based on the minimum class size.

    Args:
        min_class_size (int): The minimum class size.

    Returns:
        str: The resolved method.
    """
    if min_class_size > MAX_SET_FIT_SIZE:
        method = FINETUNING_NAME
    elif min_class_size > MAX_ANC_SET_FIT_SIZE and min_class_size <= MAX_SET_FIT_SIZE:
        method = SETFIT_NAME
    elif (
        min_class_size >= MIN_ANC_SET_FIT_SIZE
        and min_class_size <= MAX_ANC_SET_FIT_SIZE
    ):
        method = ANCSETFIT_NAME
    else:
        raise ValueError(
            f"At least 2 samples per label required for pipeline. Found: {min_class_size} sample"
        )
    return method


def count_labels_ner(dataset: Dataset) -> tuple[pd.DataFrame, int]:  # TODO: IMPLEMENT
    """
    Counts the occurrence of each label in a given dataset for NER.

    Args:
        dataset (Dataset): The input dataset.

    Returns:
        tuple[pd.DataFrame, int]: A DataFrame containing the count of each label and the minimum class size.
    """
    main_counter = Counter()
    for record in dataset:
        entities = [
            label[2:]
            for label in record["labels"]
            if label != "O" and label[:2] == "B-"
        ]
        main_counter.update(entities)
    label_counts = pd.Series(
        list(main_counter.values()), index=list(main_counter.keys())
    )
    min_class_size = label_counts.min()
    return label_counts, min_class_size


def resolve_ner_method(max_class_size: int) -> str:
    """
    Resolves the method to use for NER based on the maximum class size.

    Args:
        max_class_size (int): The maximum class size.

    Returns:
        str: The resolved method.
    """

    return TOKEN_CLF_FINETUNING



from dataclasses import dataclass
from .base_config import BaseMethodWithOODConfig
from .finetuner_config import FinetunerConfig
from ..data_types import OodMethod


@dataclass
class FinetunerOODConfig(FinetunerConfig, BaseMethodWithOODConfig):
    """Configuration for transformer finetuning with Out-of-Distribution detection.

    Extends FinetunerConfig with OOD detection using Marginal Mahalanobis distance
    by default. This method computes class-conditional Gaussian distributions in
    the embedding space and uses Mahalanobis distance to detect samples that fall
    outside known class distributions.

    This is the recommended configuration for OOD detection when sufficient
    training data is available, as Mahalanobis-based detection typically
    outperforms MSP-based methods on larger datasets.

    Attributes:
        threshold_factor: Multiplier for OOD detection threshold. Higher values
            make detection more conservative (fewer samples marked as OOD).
        ood_method: OOD detection method. Defaults to Marginal Mahalanobis,
            which estimates per-class distributions for distance-based detection.
    """

    threshold_factor: float = 1.0
    ood_method: OodMethod = OodMethod.MARGINAL_MAHALANOBIS_OOD



from dataclasses import dataclass
from .finetuner_config import FinetunerConfig


@dataclass
class TokenClassificationFinetunerConfig(FinetunerConfig):
    """Configuration for token classification (NER) finetuning."""

    pass
